import scipy as sp
import numpy as np
import scipy.ndimage as spim
from skimage.segmentation import relabel_sequential
from edt import edt
from loguru import logger
from skimage.morphology import ball, disk
from ._utils import Results
from ._unpad import unpad
try:
    from skimage.measure import marching_cubes
except ImportError:
    from skimage.measure import marching_cubes_lewiner as marching_cubes


def isolate_object(region, i, s=None):
    r"""
    Given an image containing labels, removes all labels except the specified
    one.

    Parameters
    ----------
    region : ndarray
        An image containing labelled regions, as returned by
        ``scipy.ndimage.label``.
    i : int
        The integer value
    s : tuple of slice objects, optional
        If provided, then a subsection of ``region`` will be extracted and the
        function will be applied to this subsection only.

    Returns
    -------
    label : ndarray
        An ndarray the same size as ``region`` containing *only* the objects
        with the given value ``i``.  If ``s`` is provided, the returned image
        will be a subsection of ``region``.

    """
    if s is not None:
        region = region[s]
    im = (region == i)*i
    return im


def marching_map(path, start):
    r"""
    Use the fast marching method to find distance of each voxel from a starting
    point

    Parameters
    ----------
    path : ndarray
        A boolean image with ``True`` values demarcating the path along which
        the march will occur
    start : ndarray
        A boolean image with ``True`` values indicating where the march should
        start.

    Returns
    -------
    distance : ndarray
        An array the same size as ``path`` with numerical values in each voxel
        indicating it's distance from the start point(s) along the given path.

    Notes
    -----
    This function assumes ``scikit-fmm`` is installed.

    """
    try:
        import skfmm
    except ModuleNotFoundError:
        raise ModuleNotFoundError('scikit-fmm must be install to use this ' +
                                  'function')
    phi = start*2.0 - 1.0
    speed = path*1.0
    t = skfmm.travel_time(phi, speed)
    return t.data


def align_image_with_openpnm(im):
    r"""
    Rotates an image to agree with the coordinates used in OpenPNM.

    This is necessary for overlaying the image and the network in Paraview.

    Parameters
    ----------
    im : ndarray
        The image to be rotated.  Can be the Boolean image of the pore space
        or any other image of interest.

    Returns
    -------
    image : ndarray
        Returns a copy of ``im`` rotated accordingly.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/align_image_with_openpnm.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)
    im = np.copy(im)
    if im.ndim == 2:
        im = (np.swapaxes(im, 1, 0))
        im = im[-1::-1, :]
    elif im.ndim == 3:
        im = (np.swapaxes(im, 2, 0))
        im = im[:, -1::-1, :]
    return im


def subdivide(im, divs=2, overlap=0):
    r"""
    Returns slices into an image describing the specified number of sub-arrays.

    This function is useful for performing operations on smaller images for
    memory or speed.  Note that for most typical operations this will NOT work,
    since the image borders would cause artifacts (e.g. ``distance_transform``)

    Parameters
    ----------
    im : ndarray
        The image of the porous media
    divs : scalar or array_like
        The number of sub-divisions to create in each axis of the image.  If a
        scalar is given it is assumed this value applies in all dimensions.
    overlap : scalar or array_like
        The amount of overlap to use when dividing along each axis.  If a
        scalar is given it is assumed this value applies in all dimensions.

    Returns
    -------
    slices : ndarray
        An ndarray containing sets of slice objects for indexing into ``im``
        that extract subdivisions of an image.  If ``flatten`` was ``True``,
        then this array is suitable for iterating.  If ``flatten`` was
        ``False`` then the slice objects must be accessed by row, col, layer
        indices.  An ndarray is the preferred container since its shape can
        be easily queried.

    See Also
    --------
    chunked_func

    Examples
    --------
    >>> import porespy as ps
    >>> import matplotlib.pyplot as plt
    >>> im = ps.generators.blobs(shape=[200, 200])
    >>> s = ps.tools.subdivide(im, divs=[2, 2], flatten=True)
    >>> print(len(s))
    4

    `Click here
    <https://porespy.org/examples/tools/howtos/subdivide.html>`_
    to view online example.

    """
    divs = np.ones((im.ndim,), dtype=int) * np.array(divs)
    overlap = overlap * (divs > 1)

    s = np.zeros(shape=divs, dtype=object)
    spacing = np.round(np.array(im.shape)/divs, decimals=0).astype(int)
    for i in range(s.shape[0]):
        x = spacing[0]
        sx = slice(x*i, min(im.shape[0], x*(i+1)), None)
        for j in range(s.shape[1]):
            y = spacing[1]
            sy = slice(y*j, min(im.shape[1], y*(j+1)), None)
            if im.ndim == 3:
                for k in range(s.shape[2]):
                    z = spacing[2]
                    sz = slice(z*k, min(im.shape[2], z*(k+1)), None)
                    s[i, j, k] = tuple([sx, sy, sz])
            else:
                s[i, j] = tuple([sx, sy])
    s = s.flatten().tolist()
    for i, item in enumerate(s):
        s[i] = extend_slice(slices=item, shape=im.shape, pad=overlap)
    return s


def recombine(ims, slices, overlap):
    r"""
    Recombines image chunks back into full image of original shape

    Parameters
    ----------
    ims : list of ndarrays
        The chunks of the original image, which may or may not have been
        processed.
    slices : list of slice objects
        The slice objects which were used to obtain the chunks in ``ims``
    overlap : int of list ints
        The amount of overlap used when creating chunks

    Returns
    -------
    im : ndarray
        An image constituted from the chunks in ``ims`` of the same shape
        as the original image.

    See Also
    --------
    chunked_func, subdivide

    """
    shape = [0]*ims[0].ndim
    for s in slices:
        for dim in range(len(slices[0])):
            shape[dim] = max(shape[dim], s[dim].stop)

    if isinstance(overlap, int):
        overlap = [overlap]*len(shape)

    im = np.zeros(shape, dtype=ims[0].dtype)
    for i, s in enumerate(slices):
        # Prepare new slice objects into main and sub-sliced image
        a = []  # Slices into original image
        b = []  # Slices into chunked image
        for dim in range(im.ndim):
            if s[dim].start == 0:
                ax = 0
                bx = 0
            else:
                ax = s[dim].start + overlap[dim]
                bx = overlap[dim]
            if s[dim].stop == im.shape[dim]:
                ay = im.shape[dim]
                by = im.shape[dim]
            else:
                ay = s[dim].stop - overlap[dim]
                by = s[dim].stop - s[dim].start - overlap[dim]
            a.append(slice(ax, ay, None))
            b.append(slice(bx, by, None))
        # Convert lists of slices to tuples
        a = tuple(a)
        b = tuple(b)
        # Insert image chunk into main image
        try:
            im[a] = ims[i][b]
        except ValueError:
            raise IndexError('The applied filter seems to have returned a '
                             + 'larger image that it was sent.')
    return im


def bbox_to_slices(bbox):
    r"""
    Given a tuple containing bounding box coordinates, return a tuple of slice
    objects.

    A bounding box in the form of a straight list is returned by several
    functions in skimage, but these cannot be used to direct index into an
    image.  This function returns a tuples of slices can be, such as:
    ``im[bbox_to_slices([xmin, ymin, xmax, ymax])]``.

    Parameters
    ----------
    bbox : tuple of ints
        The bounding box indices in the form (``xmin``, ``ymin``, ``zmin``,
        ``xmax``, ``ymax``, ``zmax``).  For a 2D image, simply omit the
        ``zmin`` and ``zmax`` entries.

    Returns
    -------
    slices : tuple
        A tuple of slice objects that can be used to directly index into a
        larger image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/bbox_to_slices.html>`_
    to view online example.

    """
    if len(bbox) == 4:
        ret = (slice(bbox[0], bbox[2]),
               slice(bbox[1], bbox[3]))
    else:
        ret = (slice(bbox[0], bbox[3]),
               slice(bbox[1], bbox[4]),
               slice(bbox[2], bbox[5]))
    return ret


def find_outer_region(im, r=None):
    r"""
    Find regions of the image that are outside of the solid matrix.

    Parameters
    ----------
    im : ndarray
        Image of the porous material with 1's for void and 0's for solid
    r : scalar
        The radius of the rolling ball to use.  If not specified then a value
        is calculated as twice maximum of the distance transform. The image
        size is padded by this amount in all directions, so the image can
        become quite large and unwieldy if too large a value is given.

    Returns
    -------
    image : ndarray
        A boolean mask the same shape as ``im``, containing True in all voxels
        identified as *outside* the sample.

    Notes
    -----
    This function uses the rolling ball method to define where the outer region
    ends and the void space begins.

    This is particularly useful for samples that do not fill the
    entire rectangular image, such as cylindrical cores or samples with non-
    parallel faces.

    """
    if r is None:
        dt = edt(im)
        r = int(np.amax(dt)) * 2
    im_padded = np.pad(array=im, pad_width=r, mode='constant',
                       constant_values=True)
    dt = edt(im_padded)
    seeds = (dt >= r) + get_border(shape=im_padded.shape)
    # Remove seeds not connected to edges
    labels = spim.label(seeds)[0]
    mask = labels == 1  # Assume label of 1 on edges, assured by adding border
    dt = edt(~mask)
    outer_region = dt < r
    outer_region = extract_subsection(im=outer_region, shape=im.shape)
    return outer_region


def extract_cylinder(im, r=None, axis=0):
    r"""
    Returns a cylindrical section of the image of specified radius.

    This is useful for making square images look like cylindrical cores such
    as those obtained from X-ray tomography.

    Parameters
    ----------
    im : ndarray
        The image of the porous material.  Can be any data type.
    r : scalr
        The radius of the cylinder to extract.  If ``None`` is given then the
        default is the largest cylinder that can fit inside the specified
        plane.
    axis : scalar
        The axis along with the cylinder will be oriented.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with values outside the cylindrical area set to 0 or
        ``False``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/extract_cylinder.html>`_
    to view online example.

    """
    # This needs to be imported here since the tools module is imported
    # before the generators module, so placing it at the top of the file
    # causes an error since the generators module does not exist yet.
    # Strangly, if I import the ENTIRE package at the top of the file then
    # things work ok, but this seems quite silly compared to just importing
    # the function on demand. This is explained in the following
    # stackoverflow answer: https://stackoverflow.com/a/129810.

    from porespy.generators import cylindrical_plug
    mask = cylindrical_plug(shape=im.shape, r=r, axis=axis)
    im_temp = im * mask
    return im_temp


def extract_subsection(im, shape):
    r"""
    Extracts the middle section of a image

    Parameters
    ----------
    im : ndarray
        Image from which to extract the subsection
    shape : array_like
        Can either specify the size of the extracted section or the fractional
        size of the image to extact.

    Returns
    -------
    image : ndarray
        An ndarray of size given by the ``shape`` argument, taken from the
        center of the image.

    See Also
    --------
    unpad

    Examples
    --------
    >>> import scipy as sp
    >>> from porespy.tools import extract_subsection
    >>> im = np.array([[1, 1, 1, 1], [1, 2, 2, 2], [1, 2, 3, 3], [1, 2, 3, 4]])
    >>> print(im)
    [[1 1 1 1]
     [1 2 2 2]
     [1 2 3 3]
     [1 2 3 4]]
    >>> im = extract_subsection(im=im, shape=[2, 2])
    >>> print(im)
    [[2 2]
     [2 3]]

    `Click here
    <https://porespy.org/examples/tools/howtos/extract_subsection.html>`_
    to view online example.

    """
    # Check if shape was given as a fraction
    shape = np.array(shape)
    if shape[0] < 1:
        shape = np.array(im.shape) * shape
    center = np.array(im.shape) / 2
    s_im = []
    for dim in range(im.ndim):
        r = shape[dim] / 2
        lower_im = np.amax((center[dim] - r, 0))
        upper_im = np.amin((center[dim] + r, im.shape[dim]))
        s_im.append(slice(int(lower_im), int(upper_im)))
    return im[tuple(s_im)]


def get_planes(im, squeeze=True):
    r"""
    Extracts three planar images from the volumetric image, one for each
    principle axis.  The planes are taken from the middle of the domain.

    Parameters
    ----------
    im : ndarray
        The volumetric image from which the 3 planar images are to be obtained
    squeeze : boolean, optional
        If True (default) the returned images are 2D (i.e. squeezed).  If
        False, the images are 1 element deep along the axis where the slice
        was obtained.

    Returns
    -------
    planes : list
        A list of 2D-images

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/get_planes.html>`_
    to view online example.

    """
    x, y, z = (np.array(im.shape) / 2).astype(int)
    planes = [im[x, :, :], im[:, y, :], im[:, :, z]]
    if not squeeze:
        imx = planes[0]
        planes[0] = np.reshape(imx, [1, imx.shape[0], imx.shape[1]])
        imy = planes[1]
        planes[1] = np.reshape(imy, [imy.shape[0], 1, imy.shape[1]])
        imz = planes[2]
        planes[2] = np.reshape(imz, [imz.shape[0], imz.shape[1], 1])
    return planes


def extend_slice(slices, shape, pad=1):
    r"""
    Adjust slice indices to include additional voxles around the slice.

    This function does bounds checking to ensure the indices don't extend
    outside the image.

    Parameters
    ----------
    slices : list of slice objects
         A list (or tuple) of N slice objects, where N is the number of
         dimensions in the image.
    shape : array_like
        The shape of the image into which the slice objects apply.  This is
        used to check the bounds to prevent indexing beyond the image.
    pad : int or list of ints
        The number of voxels to expand in each direction.

    Returns
    -------
    slices : list of slice objects
        A list slice of objects with the start and stop attributes respectively
        incremented and decremented by 1, without extending beyond the image
        boundaries.

    Examples
    --------
    >>> from scipy.ndimage import label, find_objects
    >>> from porespy.tools import extend_slice
    >>> im = np.array([[1, 0, 0], [1, 0, 0], [0, 0, 1]])
    >>> labels = label(im)[0]
    >>> s = find_objects(labels)

    Using the slices returned by ``find_objects``, set the first label to 3

    >>> labels[s[0]] = 3
    >>> print(labels)
    [[3 0 0]
     [3 0 0]
     [0 0 2]]

    Next extend the slice, and use it to set the values to 4

    >>> s_ext = extend_slice(s[0], shape=im.shape, pad=1)
    >>> labels[s_ext] = 4
    >>> print(labels)
    [[4 4 0]
     [4 4 0]
     [4 4 2]]

    As can be seen by the location of the 4s, the slice was extended by 1, and
    also handled the extension beyond the boundary correctly.

    """
    shape = np.array(shape)
    pad = np.array(pad).astype(int)*(shape > 0)
    a = []
    for i, s in enumerate(slices):
        start = 0
        stop = shape[i]
        start = max(s.start - pad[i], 0)
        stop = min(s.stop + pad[i], shape[i])
        a.append(slice(start, stop, None))
    return tuple(a)


def randomize_colors(im, keep_vals=[0]):
    r'''
    Takes a greyscale image and randomly shuffles the greyscale values, so that
    all voxels labeled X will be labelled Y, and all voxels labeled Y will be
    labeled Z, where X, Y, Z and so on are randomly selected from the values
    in the input image.

    This function is useful for improving the visibility of images with
    neighboring regions that are only incrementally different from each other,
    such as that returned by `scipy.ndimage.label`.

    Parameters
    ----------
    im : array_like
        An ND image of greyscale values.
    keep_vals : array_like
        Indicate which voxel values should NOT be altered.  The default is
        `[0]` which is useful for leaving the background of the image
        untouched.

    Returns
    -------
    image : ndarray
        An image the same size and type as ``im`` but with the greyscale values
        reassigned.  The unique values in both the input and output images will
        be identical.

    Notes
    -----
    If the greyscale values in the input image are not contiguous then the
    neither will they be in the output.

    Examples
    --------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> sp.random.seed(0)
    >>> im = sp.random.randint(low=0, high=5, size=[4, 4])
    >>> print(im)
    [[4 0 3 3]
     [3 1 3 2]
     [4 0 0 4]
     [2 1 0 1]]
    >>> im_rand = ps.tools.randomize_colors(im)
    >>> print(im_rand)
    [[2 0 4 4]
     [4 1 4 3]
     [2 0 0 2]
     [3 1 0 1]]

    As can be seen, the 2's have become 3, 3's have become 4, and 4's have
    become 2.  1's remained 1 by random accident.  0's remain zeros by default,
    but this can be controlled using the `keep_vals` argument.

    '''
    im_flat = im.flatten()
    keep_vals = np.array(keep_vals)
    swap_vals = ~np.in1d(im_flat, keep_vals)
    im_vals = np.unique(im_flat[swap_vals])
    new_vals = sp.random.permutation(im_vals)
    im_map = np.zeros(shape=[np.amax(im_vals) + 1, ], dtype=int)
    im_map[im_vals] = new_vals
    im_new = im_map[im_flat]
    im_new = np.reshape(im_new, newshape=np.shape(im))
    return im_new


def make_contiguous(im, mode='keep_zeros'):
    r"""
    Take an image with arbitrary greyscale values and adjust them to ensure
    all values fall in a contiguous range starting at 0.

    Parameters
    ----------
    im : array_like
        An ND array containing greyscale values
    mode : string
        Controls how the ranking is applied in the presence of numbers less
        than or equal to 0.

        'keep_zeros'
            (default) Voxels equal to 0 remain 0, and all other
            numbers are ranked starting at 1, include negative numbers,
            so [-1, 0, 4] becomes [1, 0, 2]

        'symmetric'
            Negative and positive voxels are ranked based on their
            respective distances to 0, so [-4, -1, 0, 5] becomes
            [-2, -1, 0, 1]

        'clipped'
            Voxels less than or equal to 0 are set to 0, while
            all other numbers are ranked starting at 1, so [-3, 0, 2]
            becomes [0, 0, 1].

        'none'
            Voxels are ranked such that the smallest or most
            negative number becomes 1, so [-4, 2, 0] becomes [1, 3, 2].
            This is equivalent to calling ``scipy.stats.rankdata`` directly,
            and reshaping the result to match ``im``.

    Returns
    -------
    image : ndarray
        An ndarray the same size as ``im`` but with all values in contiguous
        order.

    Examples
    --------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> im = np.array([[0, 2, 9], [6, 8, 3]])
    >>> im = ps.tools.make_contiguous(im)
    >>> print(im)
    [[0 1 5]
     [3 4 2]]

    `Click here
    <https://porespy.org/examples/tools/howtos/make_contiguous.html>`_
    to view online example.

    """
    # This is a very simple version using relabel_sequential
    im = np.array(im)
    if mode == 'none':
        im = im + np.abs(np.min(im)) + 1
        im_new = relabel_sequential(im)[0]
    if mode == 'keep_zeros':
        mask = im == 0
        im = im + np.abs(np.min(im)) + 1
        im[mask] = 0
        im_new = relabel_sequential(im)[0]
    if mode == 'clipped':
        mask = im <= 0
        im[mask] = 0
        im_new = relabel_sequential(im)[0]
    if mode == 'symmetric':
        mask = im < 0
        im_neg = relabel_sequential(-im*mask)[0]
        mask = im >= 0
        im_pos = relabel_sequential(im*mask)[0]
        im_new = im_pos - im_neg
    return im_new


def get_border(shape, thickness=1, mode='edges'):
    r"""
    Create an array with corners, edges or faces labelled as ``True``.

    This can be used as mask to manipulate values laying on the perimeter of
    an image.

    Parameters
    ----------
    shape : array_like
        The shape of the array to return.  Can be either 2D or 3D.
    thickness : scalar (default is 1)
        The number of pixels/voxels to place along perimeter.
    mode : string
        The type of border to create.  Options are 'faces', 'edges' (default)
        and 'corners'.  In 2D 'faces' and 'edges' give the same result.

    Returns
    -------
    image : ndarray
        An ndarray of specified shape with ``True`` values at the perimeter
        and ``False`` elsewhere.

    Notes
    -----
    The indices of the ``True`` values can be found using ``numpy.where``.

    Examples
    --------
    >>> import porespy as ps
    >>> import scipy as sp
    >>> mask = ps.tools.get_border(shape=[3, 3], mode='corners')
    >>> print(mask)
    [[ True False  True]
     [False False False]
     [ True False  True]]
    >>> mask = ps.tools.get_border(shape=[3, 3], mode='edges')
    >>> print(mask)
    [[ True  True  True]
     [ True False  True]
     [ True  True  True]]

    `Click here
    <https://porespy.org/examples/tools/howtos/get_border.html>`_
    to view online example.

    """
    # TODO: This function uses brute force to create an image then fills the
    # edges using location-based logic, and if the user requests
    # ``return_indices`` it finds them using ``np.where``.  Since these arrays
    # are cubic it should be possible to use more elegant and efficient
    # index-based logic to find the indices, then use them to fill an empty
    # image with ``True`` using these indices.

    ndims = len(shape)
    t = thickness
    border = np.ones(shape, dtype=bool)
    if mode == 'faces':
        if ndims == 2:
            border[t:-t, t:-t] = False
        if ndims == 3:
            border[t:-t, t:-t, t:-t] = False
    elif mode == 'edges':
        if ndims == 2:
            border[t:-t, t:-t] = False
        if ndims == 3:
            border[0::, t:-t, t:-t] = False
            border[t:-t, 0::, t:-t] = False
            border[t:-t, t:-t, 0::] = False
    elif mode == 'corners':
        if ndims == 2:
            border[t:-t, 0::] = False
            border[0::, t:-t] = False
        if ndims == 3:
            border[t:-t, 0::, 0::] = False
            border[0::, t:-t, 0::] = False
            border[0::, 0::, t:-t] = False
    return border


def in_hull(points, hull):
    """
    Test if a list of coordinates are inside a given convex hull

    Parameters
    ----------
    points : array_like (N x ndims)
        The spatial coordinates of the points to check
    hull : scipy.spatial.ConvexHull object **OR** array_like
        Can be either a convex hull object as returned by
        ``scipy.spatial.ConvexHull`` or simply the coordinates of the points
        that define the convex hull.

    Returns
    -------
    result : 1D-array
        A 1D-array Boolean array of length *N* indicating whether or not the
        given points in ``points`` lies within the provided ``hull``.

    """
    from scipy.spatial import Delaunay, ConvexHull
    if isinstance(hull, ConvexHull):
        hull = hull.points
    hull = Delaunay(hull)
    return hull.find_simplex(points) >= 0


def norm_to_uniform(im, scale=None):
    r"""
    Take an image with normally distributed greyscale values and convert it to
    a uniform (i.e. flat) distribution.

    Parameters
    ----------
    im : ndarray
        The image containing the normally distributed scalar field
    scale : [low, high]
        A list or array indicating the lower and upper bounds for the new
        randomly distributed data.  The default is ``None``, which uses the
        ``max`` and ``min`` of the original image as the the lower and upper
        bounds, but another common option might be [0, 1].

    Returns
    -------
    image : ndarray
        A copy of ``im`` with uniformly distributed greyscale values spanning
        the specified range, if given.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/norm_to_uniform.html>`_
    to view online example.

    """
    if scale is None:
        scale = [im.min(), im.max()]
    im = (im - np.mean(im)) / np.std(im)
    im = 1 / 2 * sp.special.erfc(-im / np.sqrt(2))
    im = (im - im.min()) / (im.max() - im.min())
    im = im * (scale[1] - scale[0]) + scale[0]
    return im


def _functions_to_table(mod, colwidth=[27, 48]):
    r"""
    Given a module of functions, returns a ReST formatted text string that
    outputs a table when printed.

    Parameters
    ----------
    mod : module
        The module containing the functions to be included in the table, such
        as 'porespy.filters'.
    colwidths : list of ints
        The width of the first and second columns.  Note that because of the
        vertical lines separating columns and define the edges of the table,
        the total table width will be 3 characters wider than the total sum
        of the specified column widths.

    """
    temp = mod.__dir__()
    funcs = [i for i in temp if not i[0].startswith('_')]
    funcs.sort()
    row = '+' + '-' * colwidth[0] + '+' + '-' * colwidth[1] + '+'
    fmt = '{0:1s} {1:' + str(colwidth[0] - 2) + 's} {2:1s} {3:' \
          + str(colwidth[1] - 2) + 's} {4:1s}'
    lines = []
    lines.append(row)
    lines.append(fmt.format('|', 'Method', '|', 'Description', '|'))
    lines.append(row.replace('-', '='))
    for i, item in enumerate(funcs):
        try:
            s = getattr(mod, item).__doc__.strip()
            end = s.find('\n')
            if end > colwidth[1] - 2:
                s = s[:colwidth[1] - 5] + '...'
            lines.append(fmt.format('|', item, '|', s[:end], '|'))
            lines.append(row)
        except AttributeError:
            pass
    s = '\n'.join(lines)
    return s


def mesh_region(region: bool, strel=None):
    r"""
    Creates a tri-mesh of the provided region using the marching cubes
    algorithm

    Parameters
    ----------
    im : ndarray
        A boolean image with ``True`` values indicating the region of interest
    strel : ndarray
        The structuring element to use when blurring the region.  The blur is
        perfomed using a simple convolution filter.  The point is to create a
        greyscale region to allow the marching cubes algorithm some freedom
        to conform the mesh to the surface.  As the size of ``strel`` increases
        the region will become increasingly blurred and inaccurate. The default
        is a spherical element with a radius of 1.

    Returns
    -------
    mesh : tuple
        A named-tuple containing ``faces``, ``verts``, ``norm``, and ``val``
        as returned by ``scikit-image.measure.marching_cubes`` function.

    """
    im = region
    _check_for_singleton_axes(im)
    if strel is None:
        if region.ndim == 3:
            strel = ball(1)
        if region.ndim == 2:
            strel = disk(1)
    pad_width = np.amax(strel.shape)
    if im.ndim == 3:
        padded_mask = np.pad(im, pad_width=pad_width, mode='constant')
        padded_mask = spim.convolve(padded_mask * 1.0,
                                    weights=strel) / np.sum(strel)
    else:
        padded_mask = np.reshape(im, (1,) + im.shape)
        padded_mask = np.pad(padded_mask, pad_width=pad_width, mode='constant')
    verts, faces, norm, val = marching_cubes(padded_mask)
    result = Results()
    result.verts = verts - pad_width
    result.faces = faces
    result.norm = norm
    result.val = val
    return result


def ps_disk(r, smooth=True):
    r"""
    Creates circular disk structuring element for morphological operations

    Parameters
    ----------
    r : float or int
        The desired radius of the structuring element
    smooth : boolean
        Indicates whether the faces of the sphere should have the little
        nibs (``True``) or not (``False``, default)

    Returns
    -------
    disk : ndarray
        A 2D numpy bool array of the structring element

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/ps_disk.html>`_
    to view online example.

    """
    disk = ps_round(r=r, ndim=2, smooth=smooth)
    return disk


def ps_ball(r, smooth=True):
    r"""
    Creates spherical ball structuring element for morphological operations

    Parameters
    ----------
    r : scalar
        The desired radius of the structuring element
    smooth : boolean
        Indicates whether the faces of the sphere should have the little
        nibs (``True``) or not (``False``, default)

    Returns
    -------
    ball : ndarray
        A 3D numpy array of the structuring element

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/ps_ball.html>`_
    to view online example.

    """
    ball = ps_round(r=r, ndim=3, smooth=smooth)
    return ball


def ps_round(r, ndim, smooth=True):
    r"""
    Creates round structuring element with the given radius and dimensionality

    Parameters
    ----------
    r : scalar
        The desired radius of the structuring element
    ndim : int
        The dimensionality of the element, either 2 or 3.
    smooth : boolean
        Indicates whether the faces of the sphere should have the little
        nibs (``True``) or not (``False``, default)

    Returns
    -------
    strel : ndarray
        A 3D numpy array of the structuring element

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/ps_round.html>`_
    to view online example.

    """
    rad = int(np.ceil(r))
    other = np.ones([2*rad + 1 for i in range(ndim)], dtype=bool)
    other[tuple(rad for i in range(ndim))] = False
    if smooth:
        ball = edt(other) < r
    else:
        ball = edt(other) <= r
    return ball


def ps_rect(w, ndim):
    r"""
    Creates rectilinear structuring element with the given size and
    dimensionality

    Parameters
    ----------
    w : scalar
        The desired width of the structuring element
    ndim : int
        The dimensionality of the element, either 2 or 3.

    Returns
    -------
    strel : D-aNrray
        A numpy array of the structuring element

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/ps_rect.html>`_
    to view online example.

    """
    if ndim == 2:
        from skimage.morphology import square
        strel = square(w)
    if ndim == 3:
        from skimage.morphology import cube
        strel = cube(w)
    return strel


def overlay(im1, im2, c):
    r"""
    Overlays ``im2`` onto ``im1``, given voxel coords of center of ``im2``
    in ``im1``.

    Parameters
    ----------
    im1 : ndarray
        Original voxelated image
    im2 : ndarray
        Template voxelated image
    c : array_like
        [x, y, z] coordinates in ``im1`` where ``im2`` will be centered

    Returns
    -------
    image : ndarray
        A modified version of ``im1``, with ``im2`` overlaid at the specified
        location

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/overlay.html>`_
    to view online example.

    """
    shape = im2.shape
    for ni in shape:
        if ni % 2 == 0:
            raise Exception("Structuring element must be odd-voxeled...")

    nx, ny, nz = [(ni - 1) // 2 for ni in shape]
    cx, cy, cz = c

    im1[cx - nx:cx + nx + 1, cy - ny:cy + ny + 1, cz - nz:cz + nz + 1] += im2

    return im1


def insert_sphere(im, c, r, v=True, overwrite=True):
    r"""
    Inserts a sphere of a specified radius into a given image

    Parameters
    ----------
    im : array_like
        Image into which the sphere should be inserted
    c : array_like
        The [x, y, z] coordinate indicating the center of the sphere
    r : int
        The radius of sphere to insert
    v : int
        The value to put into the sphere voxels.  The default is ``True``
        which corresponds to inserting spheres into a Boolean image.  If
        a numerical value is given, ``im`` is converted to the same type as
        ``v``.
    overwrite : boolean
        If ``True`` (default) then the sphere overwrites whatever values are
        present in ``im``.  If ``False`` then the sphere values are only
        inserted into locations that are 0 or ``False``.

    Returns
    -------
    image : ndarray
        The original image with a sphere inerted at the specified location

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/insert_sphere.html>`_
    to view online example.

    """
    # Convert image to same type os v for eventual insertion
    if im.dtype != type(v):
        im = im.astype(type(v))
    # Parse the arugments
    r = int(sp.around(r, decimals=0))
    if r == 0:
        return im
    c = np.array(c, dtype=int)
    if c.size != im.ndim:
        raise Exception('Coordinates do not match dimensionality of image')
    # Define a bounding box around inserted sphere, minding imaage boundaries
    bbox = []
    [bbox.append(np.clip(c[i] - r, 0, im.shape[i])) for i in range(im.ndim)]
    [bbox.append(np.clip(c[i] + r, 0, im.shape[i])) for i in range(im.ndim)]
    bbox = np.ravel(bbox)
    # Obtain slices into image
    s = bbox_to_slices(bbox)
    # Generate sphere template within image boundaries
    blank = np.ones_like(im[s], dtype=float)
    blank[tuple(c - bbox[0:im.ndim])] = 0.0
    sph = spim.distance_transform_edt(blank) < r
    if overwrite:  # Clear voxles under sphere to be zero
        temp = im[s] * sph > 0
        im[s][temp] = 0
    else:  # Clear portions of sphere to prevent overwriting
        sph *= im[s] == 0
    im[s] = im[s] + sph * v
    return im


def insert_cylinder(im, xyz0, xyz1, r):
    r"""
    Inserts a cylinder of given radius onto an image

    Parameters
    ----------
    im : array_like
        Original voxelated image
    xyz0, xyz1 : 3-by-1 array_like
        Voxel coordinates of the two end points of the cylinder
    r : int
        Radius of the cylinder

    Returns
    -------
    im : ndarray
        Original voxelated image overlayed with the cylinder

    Notes
    -----
    This function is only implemented for 3D images

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/insert_cylinder.html>`_
    to view online example.

    """
    if im.ndim != 3:
        raise Exception('This function is only implemented for 3D images')
    # Converting coordinates to numpy array
    xyz0, xyz1 = [np.array(xyz).astype(int) for xyz in (xyz0, xyz1)]
    r = int(r)
    L = np.absolute(xyz0 - xyz1).max() + 1
    xyz_line = [np.linspace(xyz0[i], xyz1[i], L).astype(int) for i in range(3)]

    for i, c in enumerate(xyz_line):
        if c.min() < 0:
            raise Exception('Given endpoint coordinates lie outside image')
        if c.max() > im.shape[i]:
            raise Exception('Given endpoint coordinates lie outside image')
        c += r

    im = np.pad(im, r)
    xyz_min = np.amin(xyz_line, axis=1) - r
    xyz_max = np.amax(xyz_line, axis=1) + r
    shape_template = xyz_max - xyz_min + 1
    template = np.zeros(shape=shape_template)

    # Shortcut for orthogonal cylinders
    if (xyz0 == xyz1).sum() == 2:
        unique_dim = [xyz0[i] != xyz1[i] for i in range(3)].index(True)
        shape_template[unique_dim] = 1
        template_2D = disk(radius=r).reshape(shape_template)
        template = np.repeat(template_2D, repeats=L, axis=unique_dim)
        xyz_min[unique_dim] += r
        xyz_max[unique_dim] += -r
    else:
        xyz_line_in_template_coords = [xyz_line[i] - xyz_min[i] for i in range(3)]
        template[tuple(xyz_line_in_template_coords)] = 1
        template = edt(template == 0) <= r

    im[xyz_min[0]: xyz_max[0] + 1,
       xyz_min[1]: xyz_max[1] + 1,
       xyz_min[2]: xyz_max[2] + 1] += template

    im = unpad(im, r)

    return im


def extract_regions(regions, labels: list, trim=True):
    r"""
    Combine given regions into a single boolean mask

    Parameters
    -----------
    regions : ndarray
        An image containing an arbitrary number of labeled regions
    labels : array_like or scalar
        A list of labels indicating which region or regions to extract
    trim : bool
        If ``True`` then image shape will trimmed to a bounding box around the
        given regions.

    Returns
    -------
    im : ndarray
        A boolean mask with ``True`` values indicating where the given labels
        exist

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/extract_regions.html>`_
    to view online example.

    """
    if type(labels) is int:
        labels = [labels]
    s = spim.find_objects(regions)
    im_new = np.zeros_like(regions)
    x_min, y_min, z_min = sp.inf, sp.inf, sp.inf
    x_max, y_max, z_max = 0, 0, 0
    for i in labels:
        im_new[s[i - 1]] = regions[s[i - 1]] == i
        x_min, x_max = min(s[i - 1][0].start, x_min), max(s[i - 1][0].stop, x_max)
        y_min, y_max = min(s[i - 1][1].start, y_min), max(s[i - 1][1].stop, y_max)
        if regions.ndim == 3:
            z_min, z_max = min(s[i - 1][2].start, z_min), max(s[i - 1][2].stop, z_max)
    if trim:
        if regions.ndim == 3:
            bbox = bbox_to_slices([x_min, y_min, z_min, x_max, y_max, z_max])
        else:
            bbox = bbox_to_slices([x_min, y_min, x_max, y_max])
        im_new = im_new[bbox]
    return im_new


def _check_for_singleton_axes(im):  # pragma: no cover
    r"""
    Checks for whether the input image contains singleton axes and logs
    a proper warning in case found.

    Parameters
    ----------
    im : ndarray
        Input image.

    """
    if im.ndim != im.squeeze().ndim:
        logger.warning("Input image conains a singleton axis. Reduce"
                       " dimensionality with np.squeeze(im) to avoid"
                       " unexpected behavior.")



import numpy as np
from porespy.networks import regions_to_network
from porespy.networks import add_boundary_regions
from porespy.networks import label_phases, label_boundaries
from porespy.filters import snow_partitioning, snow_partitioning_parallel
from porespy.tools import Results
from loguru import logger


def snow2(phases,
          phase_alias=None,
          boundary_width=3,
          accuracy='standard',
          voxel_size=1,
          sigma=0.4,
          r_max=4,
          parallelization={},):
    r"""
    Applies the SNOW algorithm to each phase indicated in ``phases``.

    This function is a combination of ``snow`` [1]_, ``snow_dual`` [2]_,
    ``snow_n`` [3]_, and ``snow_parallel`` [4]_ from previous versions.

    Parameters
    ----------
    phases : ndarray
        An image indicating the phase(s) of interest. A watershed is
        produced for each integer value in ``phases`` (except 0's). These
        are then combined into a single image and one network is extracted
        using ``regions_to_network``.
    phase_alias : dict
        A mapping between integer values in ``phases`` and phase name
        used to add labels to the network. For instance, asssuming a
        two-phase image, ``{1: 'void', 2: 'solid'}`` will result in the
        labels ``'pore.void'`` and ``'pore.solid'``, as well as
        ``'throat.solid_void'``, ``'throat.solid_solid'``, and
        ``'throat.void_void'``. If not provided, aliases are assumed to be
        ``{1: 'phase1', 2: 'phase2, ...}``.  Phase labels can also be
        applied afterward using ``label_phases``.
    boundary_width : depends
        Number of voxels to add to the beginning and end of each axis.
        This argument can either be a scalar or a list. If a scalar is
        passed, it will be applied to the beginning and end of all axes.
        In case of a list, you can specify the number of voxels for each
        axis individually. Here are some examples:

            - [0, 3, 0]: 3 voxels only applied to the y-axis.

            - [0, [0, 3], 0]: 3 voxels only applied to the end of y-axis.

            - [0, [3, 0], 0]: 3 voxels only applied to the beginning of y-axis.

        The default is to add 3 voxels on both ends of all axes. For each
        boundary width that is not 0, a label will automatically be
        applied indicating which end of which axis (i.e. ``'xmin'`` and
        ``'xmax'``).
    accuracy : string
        Controls how accurately certain properties are calculated during
        the analysis of regions in the ``regions_to_network`` function.
        Options are:

            - 'standard' (default)
                Computes the surface areas and perimeters by simply
                counting voxels. This is *much* faster but does not
                properly account for the rough voxelated nature
                of the surfaces.

            - 'high'
                Computes surface areas using the marching cube
                method, and perimeters using the fast marching method. These
                are substantially slower but better account for the
                voxelated nature of the images.

    voxel_size : scalar (default = 1)
        The resolution of the image, expressed as the length of one side
        of a voxel, so the volume of a voxel would be **voxel_size**-cubed.
    r_max : int
        The radius of the spherical structuring element to use in the
        Maximum filter stage that is used to find peaks. The default is 4.
    sigma : float
        The standard deviation of the Gaussian filter used in step 1. The
        default is 0.4.  If 0 is given then the filter is not applied.
    parallelization : dict
        The arguments for controlling the parallelization of the watershed
        function are rolled into this dictionary, otherwise the function
        signature would become too complex. Refer to the docstring of
        ``snow_partitioning_parallel`` for complete details. If no values
        are provided then the defaults for that function are used here.
        To disable parallelization pass ``parallel=None``, which will
        invoke the standard ``snow_partitioning`` or ``snow_partitioning_n``.

    Returns
    -------
    network : Results object
        A custom object is returned with the following data added as attributes:

        - 'phases'
            The original ``phases`` image with any padding applied

        - 'regions'
            The watershed segmentation of the image, including boundary
            regions if padding was applied

        - 'network'
            A dictionary containing all the extracted network properties in
            OpenPNM format ('pore.coords', 'throat.conns', etc).

    References
    ----------
    .. [1] Gostick JT. Versatile and efficient pore network extraction
       method using marker-based watershed segmentation. Phys. Rev. E. 96,
       023307 (2017)
    .. [2] Khan ZA, Tranter TG, Agnaou M, Elkamel A, and Gostick JT, Dual
       network extraction algorithm to investigate multiple transport
       processes in porous materials: Image-based modeling of pore and
       grain-scale processes. Computers and Chemical Engineering. 123(6),
       64-77 (2019)
    .. [3] Khan ZA, García-Salaberri PA, Heenan T, Jervis R, Shearing P,
       Brett D, Elkamel A, Gostick JT, Probing the structure-performance
       relationship of lithium-ion battery cathodes using pore-networks
       extracted from three-phase tomograms. Journal of the
       Electrochemical Society. 167(4), 040528 (2020)
    .. [4] Khan ZA, Elkamel A, Gostick JT, Efficient extraction of pore
       networks from massive tomograms via geometric domain decomposition.
       Advances in Water Resources. 145(Nov), 103734 (2020)

    """
    regions = None
    for i in range(phases.max()):
        phase = phases == (i + 1)
        if parallelization is not None:
            snow = snow_partitioning_parallel(
                im=phase, sigma=sigma, r_max=r_max, **parallelization)
        else:
            snow = snow_partitioning(im=phase, sigma=sigma, r_max=r_max)
        if regions is None:
            regions = np.zeros_like(snow.regions, dtype=int)
        # Note: Using snow.regions > 0 here instead of phase is needed to
        # handle a bug in snow_partitioning, see issue #169 and #430
        regions += snow.regions + regions.max()*(snow.regions > 0)
    if phases.shape != regions.shape:
        logger.warning(f"Image was cropped to {regions.shape} during watershed")
        for ax in range(phases.ndim):
            phases = np.swapaxes(phases, 0, ax)
            phases = phases[:regions.shape[ax], ...]
            phases = np.swapaxes(phases, 0, ax)
    # Inspect and clean-up boundary_width argument
    boundary_width = _parse_pad_width(boundary_width, phases.shape)
    # If boundaries were specified, pad the images accordingly
    if np.any(boundary_width):
        regions = add_boundary_regions(regions, pad_width=boundary_width)
        phases = np.pad(phases, pad_width=boundary_width, mode='edge')
    # Perform actual extractcion on all regions
    net = regions_to_network(
        regions, phases=phases, accuracy=accuracy, voxel_size=voxel_size)
    # If image is multiphase, label pores/throats accordingly
    if phases.max() > 1:
        phase_alias = _parse_phase_alias(phase_alias, phases)
        net = label_phases(net, alias=phase_alias)
    # If boundaries were added, label them accordingly
    if np.any(boundary_width):
        W = boundary_width.flatten()
        L = ['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax'][:phases.ndim*2]
        L = [L[i]*int(W[i] > 0) for i in range(len(L))]
        L = np.reshape(L, newshape=boundary_width.shape)
        net = label_boundaries(net, labels=L)
    result = Results()
    result.network = net
    result.regions = regions
    result.phases = phases
    return result


def _parse_phase_alias(alias, phases):
    r"""
    """
    if alias is None:
        alias = {i+1: 'phase' + str(i+1) for i in range(phases.max())}
    for i in range(phases.max()):
        if i+1 not in alias.keys():
            alias[i+1] = 'phase'+str(i+1)
    return alias


def _parse_pad_width(pad_width, shape):
    r"""
    """
    ndim = len(shape)
    pad_width = np.atleast_1d(np.array(pad_width, dtype=object))

    if np.size(pad_width) == 1:
        pad_width = np.tile(pad_width.item(), ndim).astype(object)
    if len(pad_width) != ndim:
        raise Exception(f"pad_width must be scalar or {ndim}-element list")

    tmp = []
    for elem in pad_width:
        if np.size(elem) == 1:
            tmp.append(np.tile(np.array(elem).item(), 2))
        elif np.size(elem) == 2 and np.ndim(elem) == 1:
            tmp.append(elem)
        else:
            raise Exception("pad_width components can't have 2+ elements")

    return np.array(tmp)



import numpy as np
import scipy.ndimage as spim
from porespy.tools import get_tqdm
from porespy import settings
from loguru import logger
tqdm = get_tqdm()


def random_cantor_dust(shape, n, p=2, f=0.8):
    r"""
    Generates an image of random cantor dust

    Parameters
    ----------
    shape : array_like
        The shape of the final image.  If not evenly divisible by $p**n$
        it will be increased to the nearest size that is.
    n : int
        The number of times to iteratively divide the image.
    p : int (default = 2)
        The number of divisions to make on each iteration.
    f : float (default = 0.8)
        The fraction of the set to keep on each iteration.

    Returns
    -------
    dust : ndarray
        A boolean image of a random Cantor dust

    """
    # Parse the given shape and adjust if necessary
    shape = np.array(shape)
    trim = np.mod(shape, (p**n))
    if np.any(trim > 0):
        shape = shape - trim + p**n
        logger.warning(f"Requested shape being changed to {shape}")
    im = np.ones(shape, dtype=bool)
    divs = []
    if isinstance(n, int):
        for i in range(1, n):
            divs.append(p**i)
    else:
        for i in n:
            divs.append(p**i)
    for i in tqdm(divs, **settings.tqdm):
        sh = (np.array(im.shape)/i).astype(int)
        mask = np.random.rand(*sh) < f
        mask = spim.zoom(mask, zoom=i, order=0)
        im = im*mask
    return im


def sierpinski_foam(dmin, n, ndim=2, max_size=1e9):
    r"""
    Generates an image of a Sierpinski carpet or foam

    Parameters
    ----------
    dmin : int
        The size of the smallest square in the final image
    n : int
        The number of times to iteratively tile the image
    ndim : int
        The number of dimensions of the desired image, can be 2 or 3. The
        default value is 2.

    Returns
    -------
    foam : ndarray
        A boolean image of a Sierpinski gasket or foam

    """
    def _insert_cubes(im, n):
        if n > 0:
            n -= 1
            shape = np.asarray(np.shape(im))
            im = np.tile(im, (3, 3, 3))
            im[shape[0]:2*shape[0], shape[1]:2*shape[1], shape[2]:2*shape[2]] = 0
            if im.size < max_size:
                im = _insert_cubes(im, n)
        return im

    def _insert_squares(im, n):
        if n > 0:
            n -= 1
            shape = np.asarray(np.shape(im))
            im = np.tile(im, (3, 3))
            im[shape[0]:2*shape[0], shape[1]:2*shape[1]] = 0
            if im.size < max_size:
                im = _insert_squares(im, n)
        return im

    im = np.ones([dmin]*ndim, dtype=int)
    if ndim == 2:
        im = _insert_squares(im, n)
    elif ndim == 3:
        im = _insert_cubes(im, n)
    return im



import inspect as insp
import dask
import numpy as np
from edt import edt
import operator as op
import scipy.ndimage as spim
from skimage.morphology import reconstruction
from skimage.segmentation import clear_border
from skimage.morphology import ball, disk, square, cube, diamond, octahedron
from porespy.tools import _check_for_singleton_axes
from porespy.tools import get_border, subdivide, recombine
from porespy.tools import unpad, extract_subsection
from porespy.tools import ps_disk, ps_ball
from porespy import settings
from porespy.tools import get_tqdm
from loguru import logger
tqdm = get_tqdm()


def apply_padded(im, pad_width, func, pad_val=1, **kwargs):
    r"""
    Applies padding to an image before sending to ``func``, then extracts
    the result corresponding to the original image shape.

    Parameters
    ----------
    im : ndarray
        The image to which ``func`` should be applied
    pad_width : int or list of ints
        The amount of padding to apply to each axis. Refer to
        ``numpy.pad`` documentation for more details.
    pad_val : scalar
        The value to place into the padded voxels.  The default is 1 (or
        ``True``) which extends the pore space.
    func : function handle
        The function to apply to the padded image.
    kwargs
        Additional keyword arguments are collected and passed to ``func``.

    Notes
    -----
    A use case for this is when using ``skimage.morphology.skeletonize_3d``
    to ensure that the skeleton extends beyond the edges of the image.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/apply_padded.html>`_
    to view online example.

    """
    padded = np.pad(im, pad_width=pad_width,
                    mode='constant', constant_values=pad_val)
    temp = func(padded, **kwargs)
    result = unpad(im=temp, pad_width=pad_width)
    return result


def trim_small_clusters(im, size=1):
    r"""
    Remove isolated voxels or clusters of a given size or smaller

    Parameters
    ----------
    im : ndarray
        The binary image from which voxels are to be removed.
    size : scalar
        The threshold size of clusters to trim.  As clusters with this
        many voxels or fewer will be trimmed.  The default is 1 so only
        single voxels are removed.

    Returns
    -------
    im : ndarray
        A copy of ``im`` with clusters of voxels smaller than the given
        ``size`` removed.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/trim_small_clusters.html>`_
    to view online example.

    """
    if im.ndim == 2:
        strel = disk(1)
    elif im.ndim == 3:
        strel = ball(1)
    else:
        raise Exception("Only 2D or 3D images are accepted")
    filtered_array = np.copy(im)
    labels, N = spim.label(filtered_array, structure=strel)
    id_sizes = np.array(spim.sum(im, labels, range(N + 1)))
    area_mask = id_sizes <= size
    filtered_array[area_mask[labels]] = 0
    return filtered_array


def hold_peaks(im, axis=-1):
    r"""
    Replaces each voxel with the highest value along the given axis.

    Parameters
    ----------
    im : ndarray
        A greyscale image whose peaks are to be found.
    axis : int
        The axis along which the operation is to be applied.

    Returns
    -------
    result : ndarray
        A copy of ``im`` with each voxel is replaced with the highest value along
        the given axis.

    Notes
    -----
    "im" must be a greyscale image. In case a Boolean image is fed into this
    method, it will be first converted to float values [0.0,1.0] before proceeding.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/hold_peaks.html>`_
    to view online example.

    """

    A = im.astype(float)
    B = np.swapaxes(A, axis, -1)
    updown = np.empty((*B.shape[:-1], B.shape[-1] + 1), B.dtype)
    updown[..., 0], updown[..., -1] = -1, -1
    np.subtract(B[..., 1:], B[..., :-1], out=updown[..., 1:-1])
    chnidx = np.where(updown)
    chng = updown[chnidx]
    (pkidx,) = np.where((chng[:-1] > 0) & (chng[1:] < 0) | (chnidx[-1][:-1] == 0))
    pkidx = (*map(op.itemgetter(pkidx), chnidx),)
    out = np.zeros_like(A)
    aux = out.swapaxes(axis, -1)
    aux[(*map(op.itemgetter(slice(1, None)), pkidx),)] = np.diff(B[pkidx])
    aux[..., 0] = B[..., 0]
    result = out.cumsum(axis=axis)
    return result


def distance_transform_lin(im, axis=0, mode="both"):
    r"""
    Replaces each void voxel with the linear distance to the nearest solid
    voxel along the specified axis.

    Parameters
    ----------
    im : ndarray
        The image of the porous material with ``True`` values indicating
        the void phase (or phase of interest).
    axis : int
        The direction along which the distance should be measured, the
        default is 0 (i.e. along the x-direction).
    mode : str
        Controls how the distance is measured. Options are:

        'forward'
            Distances are measured in the increasing direction
            along the specified axis
        'reverse'
            Distances are measured in the reverse direction.
            'backward' is also accepted.
        'both'
            Distances are calculated in both directions (by
            recursively calling itself), then reporting the minimum value
            of the two results.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with each foreground voxel containing the
        distance to the nearest background along the specified axis.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/distance_transform_lin.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)

    if mode in ["backward", "reverse"]:
        im = np.flip(im, axis)
        im = distance_transform_lin(im=im, axis=axis, mode="forward")
        im = np.flip(im, axis)
        return im
    elif mode in ["both"]:
        im_f = distance_transform_lin(im=im, axis=axis, mode="forward")
        im_b = distance_transform_lin(im=im, axis=axis, mode="backward")
        return np.minimum(im_f, im_b)
    b = np.cumsum(im > 0, axis=axis)
    c = np.diff(b * (im == 0), axis=axis)
    d = np.minimum.accumulate(c, axis=axis)
    if im.ndim == 1:
        e = np.pad(d, pad_width=[1, 0], mode="constant", constant_values=0)
    elif im.ndim == 2:
        ax = [[[1, 0], [0, 0]], [[0, 0], [1, 0]]]
        e = np.pad(d, pad_width=ax[axis], mode="constant", constant_values=0)
    elif im.ndim == 3:
        ax = [
            [[1, 0], [0, 0], [0, 0]],
            [[0, 0], [1, 0], [0, 0]],
            [[0, 0], [0, 0], [1, 0]],
        ]
        e = np.pad(d, pad_width=ax[axis], mode="constant", constant_values=0)
    f = im * (b + e)
    return f


def find_disconnected_voxels(im, conn=None):
    r"""
    Identifies all voxels that are not connected to the edge of the image.

    Parameters
    ----------
    im : ndarray
        A Boolean image, with ``True`` values indicating the phase for which
        disconnected voxels are sought.
    conn : int
        For 2D the options are 4 and 8 for square and diagonal neighbors,
        while for the 3D the options are 6 and 26, similarily for square
        and diagonal neighbors. The default is the maximum option.

    Returns
    -------
    image : ndarray
        An ndarray the same size as ``im``, with True values indicating
        voxels of the phase of interest (i.e. True values in the original
        image) that are not connected to the outer edges.

    See Also
    --------
    fill_blind_pores, trim_floating_solid

    Notes
    -----
    This function is just a convenient wrapper around the ``clear_border``
    function of ``scikit-image``.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/find_disconnected_voxels.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)

    if im.ndim == 2:
        if conn == 4:
            strel = disk(1)
        elif conn in [None, 8]:
            strel = square(3)
        else:
            raise Exception("Received conn is not valid")
    elif im.ndim == 3:
        if conn == 6:
            strel = ball(1)
        elif conn in [None, 26]:
            strel = cube(3)
        else:
            raise Exception("Received conn is not valid")
    labels, N = spim.label(input=im, structure=strel)
    holes = clear_border(labels=labels) > 0
    return holes


def fill_blind_pores(im, conn=None):
    r"""
    Fills all pores that are not connected to the edges of the image.

    Parameters
    ----------
    im : ndarray
        The image of the porous material

    Returns
    -------
    im : ndarray
        A version of ``im`` but with all the disconnected pores removed.
    conn : int
        For 2D the options are 4 and 8 for square and diagonal neighbors,
        while for the 3D the options are 6 and 26, similarily for square
        and diagonal neighbors. The default is the maximum option.

    See Also
    --------
    find_disconnected_voxels

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/fill_blind_pores.html>`_
    to view online example.

    """
    im = np.copy(im)
    holes = find_disconnected_voxels(im, conn=conn)
    im[holes] = False
    return im


def trim_floating_solid(im, conn=None):
    r"""
    Removes all solid that that is not attached to the edges of the image.

    Parameters
    ----------
    im : ndarray
        The image of the porous material
    conn : int
        For 2D the options are 4 and 8 for square and diagonal neighbors,
        while for the 3D the options are 6 and 26, similarily for square
        and diagonal neighbors. The default is the maximum option.

    Returns
    -------
    image : ndarray
        A version of ``im`` but with all the disconnected solid removed.

    See Also
    --------
    find_disconnected_voxels

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/trim_floating_solid.html>`_
    to view online example.

    """
    im = np.copy(im)
    holes = find_disconnected_voxels(~im, conn=conn)
    im[holes] = True
    return im


def trim_nonpercolating_paths(im, inlets, outlets):
    r"""
    Remove all nonpercolating paths between specified locations

    Parameters
    ----------
    im : ndarray
        The image of the porous material with ```True`` values indicating the
        phase of interest
    inlets : ndarray
        A boolean mask indicating locations of inlets, such as produced by
        ``porespy.generators.faces``.
    outlets : ndarray
        A boolean mask indicating locations of outlets, such as produced by
        ``porespy.generators.faces``.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with all the nonpercolating paths removed

    Notes
    -----
    This function is essential when performing transport simulations on an
    image since regions that do not span between the desired inlet and
    outlet do not contribute to the transport.

    See Also
    --------
    find_disconnected_voxels
    trim_floating_solid
    trim_blind_pores

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/trim_nonpercolating_paths.html>`_
    to view online example.

    """
    labels = spim.label(im)[0]
    IN = np.unique(labels * inlets)
    OUT = np.unique(labels * outlets)
    hits = np.array(list(set(IN).intersection(set(OUT))))
    new_im = np.isin(labels, hits[hits > 0])
    return new_im


def trim_extrema(im, h, mode="maxima"):
    r"""
    Trims local extrema in greyscale values by a specified amount.

    This essentially decapitates peaks and/or floods valleys.

    Parameters
    ----------
    im : ndarray
        The image whose extrema are to be removed
    h : float
        The height to remove from each peak or fill in each valley
    mode : string {'maxima' | 'minima' | 'extrema'}
        Specifies whether to remove maxima or minima or both

    Returns
    -------
    image : ndarray
        A copy of the input image with all the peaks and/or valleys
        removed.

    Notes
    -----
    (1) This function is referred to as **imhmax** or **imhmin** in Matlab.

    (2) If the provided ``h`` is larger than ALL peaks in the array, then the
    baseline values of the array are changed as well.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/trim_extrema.html>`_
    to view online example.

    """
    mask = np.copy(im)
    im = np.copy(im)
    if mode == 'maxima':
        result = reconstruction(seed=im - h, mask=mask, method='dilation')
    elif mode == 'minima':
        result = reconstruction(seed=im + h, mask=mask, method='erosion')
    elif mode == 'extrema':
        result = reconstruction(seed=im - h, mask=mask, method='dilation')
        result = reconstruction(seed=result + h, mask=result, method='erosion')
    return result


def flood(im, labels, mode="max"):
    r"""
    Floods/fills each region in an image with a single value based on the
    specific values in that region.

    This function calls the various functions in ``scipy.ndimage.measurements``
    but instead of returning a list of values, it fills each region with its
    value.  This is useful for visualization and statistics.

    Parameters
    ----------
    im : array_like
        An image with the numerical values of interest in each voxel,
        and 0's elsewhere.
    labels : array_like
        An array the same shape as ``im`` with each region labeled.
    mode : string
        Specifies how to determine the value to flood each region. Options
        taken from the ``scipy.ndimage.measurements`` functions:

            'maximum'
                Floods each region with the local max in that region. The
                keyword ``max`` is also accepted.
            'minimum'
                Floods each region the local minimum in that region. The
                keyword ``min`` is also accepted.
            'median'
                Floods each region the local median in that region
            'mean'
                Floods each region the local mean in that region
            'size'
                Floods each region with the size of that region.  This is
                actually accomplished with ``scipy.ndimage.sum`` by converting
                ``im`` to a boolean image (``im = im > 0``).
            'standard_deviation'
                Floods each region with the value of the standard deviation
                of the voxels in ``im``.
            'variance'
                Floods each region with the value of the variance of the voxels
                in ``im``.

    Returns
    -------
    flooded : ndarray
        A copy of ``im`` with new values placed in each forground voxel
        based on the ``mode``.

    See Also
    --------
    prop_to_image, flood_func, region_size

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/flood.html>`_
    to view online example.

    """
    mask = im > 0
    N = labels.max()
    mode = "sum" if mode == "size" else mode
    mode = "maximum" if mode == "max" else mode
    mode = "minimum" if mode == "min" else mode
    f = getattr(spim, mode)
    vals = f(input=im, labels=labels, index=range(0, N + 1))
    flooded = vals[labels]
    flooded = flooded * mask
    return flooded


def flood_func(im, labels, func):
    r"""
    Flood each isolated region in an image with a constant value calculated by
    the given function.

    Parameters
    ----------
    im : ndarray
        An image with the numerical values of interest in each voxel,
        and 0's elsewhere.
    labels : ndarray
        An array containing labels identify each individual region to be
        flooded. If not provided then ``scipy.ndimage.label`` is applied to
        ``im > 0``.
    func : Numpy function handle
        The function to be applied to each region in the image.  Any Numpy
        function that returns a scalar value can be passed, such as ``amin``,
        ``amax``, ``sum``, ``mean``, ``median``, etc.

    Returns
    -------
    flooded : ndarray
        An image the same size as ``im`` with each isolated region flooded
        with a constant value based on the given ``func`` and the values
        in ``im``.

    See Also
    --------
    flood, region_size

    Notes
    -----
    Many of the functions in ``scipy.ndimage`` can be applied to
    individual regions using the ``index`` argument.  This function extend
    that behavior to all numpy function, in the event you wanted to compute
    the cosine of the values in each region for some reason. This function
    also floods the original image instead of return a list of values for
    each region.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/flood_func.html>`_
    to view online example.

    """
    slices = spim.find_objects(labels)
    flooded = np.zeros_like(im, dtype=float)
    for i, s in enumerate(slices):
        sub_im = labels[s] == (i + 1)
        val = func(im[s][sub_im])
        flooded[s] += sub_im*val
    return flooded


def find_dt_artifacts(dt):
    r"""
    Label points in a distance transform that are closer to image boundary
    than solid

    These points could *potentially* be erroneously high since their
    distance values do not reflect the possibility that solid may have
    been present beyond the border of the image but was lost by trimming.

    Parameters
    ----------
    dt : ndarray
        The distance transform of the phase of interest.

    Returns
    -------
    image : ndarray
        An ndarray the same shape as ``dt`` with numerical values
        indicating the maximum amount of error in each volxel, which is
        found by subtracting the distance to nearest edge of image from
        the distance transform value. In other words, this is the error
        that would be found if there were a solid voxel lurking just
        beyond the nearest edge of the image.  Obviously, voxels with a
        value of zero have no error.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/find_dt_artifacts.html>`_
    to view online example.

    """
    temp = np.ones(shape=dt.shape) * np.inf
    for ax in range(dt.ndim):
        dt_lin = distance_transform_lin(np.ones_like(temp, dtype=bool),
                                        axis=ax, mode="both")
        temp = np.minimum(temp, dt_lin)
    result = np.clip(dt - temp, a_min=0, a_max=np.inf)
    return result


def region_size(im):
    r"""
    Replace each voxel with the size of the region to which it belongs

    Parameters
    ----------
    im : ndarray
        Either a boolean image wtih ``True`` indicating the features of
        interest, in which case ``scipy.ndimage.label`` will be applied to
        find regions, or a greyscale image with integer values indicating
        regions.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with each voxel value indicating the size of the
        region to which it belongs.  This is particularly useful for
        finding chord sizes on the image produced by ``apply_chords``.

    See Also
    --------
    flood

    Notes
    -----
    This function provides the same result as ``flood`` with ``mode='size'``,
    although does the computation in a different way.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/region_size.html>`_
    to view online example.

    """
    if im.dtype == bool:
        im = spim.label(im)[0]
    counts = np.bincount(im.flatten())
    counts[0] = 0
    return counts[im]


def apply_chords(im, spacing=1, axis=0, trim_edges=True, label=False):
    r"""
    Adds chords to the void space in the specified direction.

    Parameters
    ----------
    im : ndarray
        An image of the porous material with void marked as ``True``.
    spacing : int
        Separation between chords.  The default is 1 voxel.  This can be
        decreased to 0, meaning that the chords all touch each other,
        which automatically sets to the ``label`` argument to ``True``.
    axis : int (default = 0)
        The axis along which the chords are drawn.
    trim_edges : bool (default = ``True``)
        Whether or not to remove chords that touch the edges of the image.
        These chords are artifically shortened, so skew the chord length
        distribution.
    label : bool (default is ``False``)
        If ``True`` the chords in the returned image are each given a
        unique label, such that all voxels lying on the same chord have
        the same value.  This is automatically set to ``True`` if spacing
        is 0, but is ``False`` otherwise.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with non-zero values indicating the chords.

    See Also
    --------
    apply_chords_3D

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/apply_chords.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)
    if spacing < 0:
        raise Exception("Spacing cannot be less than 0")
    if spacing == 0:
        label = True
    result = np.zeros(im.shape, dtype=int)  # Will receive chords at end
    slxyz = [slice(None, None, spacing * (axis != i) + 1) for i in [0, 1, 2]]
    slices = tuple(slxyz[: im.ndim])
    s = [[0, 1, 0], [0, 1, 0], [0, 1, 0]]  # Straight-line structuring element
    if im.ndim == 3:  # Make structuring element 3D if necessary
        s = np.pad(np.atleast_3d(s), pad_width=((0, 0), (0, 0), (1, 1)),
                   mode="constant", constant_values=0)
    im = im[slices]
    s = np.swapaxes(s, 0, axis)
    chords = spim.label(im, structure=s)[0]
    if trim_edges:  # Label on border chords will be set to 0
        chords = clear_border(chords)
    result[slices] = chords  # Place chords into empty image created at top
    if label is False:  # Remove label if not requested
        result = result > 0
    return result


def apply_chords_3D(im, spacing=0, trim_edges=True):
    r"""
    Adds chords to the void space in all three principle directions.

    Chords in the X, Y and Z directions are labelled 1, 2 and 3 resepctively.

    Parameters
    ----------
    im : ndarray
        A 3D image of the porous material with void space marked as True.
    spacing : int (default = 0)
        Chords are automatically separed by 1 voxel on all sides, and this
        argument increases the separation.
    trim_edges : bool (default is ``True``)
        Whether or not to remove chords that touch the edges of the image.
        These chords are artifically shortened, so skew the chord length
        distribution

    Returns
    -------
    image : ndarray
        A copy of ``im`` with values of 1 indicating x-direction chords,
        2 indicating y-direction chords, and 3 indicating z-direction
        chords.

    Notes
    -----
    The chords are separated by a spacing of at least 1 voxel so that
    tools that search for connected components, such as
    ``scipy.ndimage.label`` can detect individual chords.

    See Also
    --------
    apply_chords

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/apply_chords_3D.html>`_
    to view online example.

    """
    _check_for_singleton_axes(im)
    if im.ndim < 3:
        raise Exception("Must be a 3D image to use this function")
    if spacing < 0:
        raise Exception("Spacing cannot be less than 0")
    ch = np.zeros_like(im, dtype=int)
    ch[:, :: 4 + 2 * spacing, :: 4 + 2 * spacing] = 1   # X-direction
    ch[:: 4 + 2 * spacing, :, 2::4 + 2 * spacing] = 2   # Y-direction
    ch[2::4 + 2 * spacing, 2::4 + 2 * spacing, :] = 3   # Z-direction
    chords = ch * im
    if trim_edges:
        temp = clear_border(spim.label(chords > 0)[0]) > 0
        chords = temp * chords
    return chords


def local_thickness(im, sizes=25, mode="hybrid", divs=1):
    r"""
    For each voxel, this function calculates the radius of the largest
    sphere that both engulfs the voxel and fits entirely within the
    foreground.

    This is not the same as a simple distance transform, which finds the
    largest sphere that could be *centered* on each voxel.

    Parameters
    ----------
    im : ndarray
        A binary image with the phase of interest set to True
    sizes : array_like or scalar
        The sizes to invade.  If a list of values of provided they are
        used directly. If a scalar is provided then that number of points
        spanning the min and max of the distance transform are used.
    mode : str
        Controls with method is used to compute the result. Options are:

        'hybrid'
            (default) Performs a distance transform of the void
            space, thresholds to find voxels larger than ``sizes[i]``, trims
            the resulting mask if ``access_limitations`` is ``True``, then
            dilates it using the efficient fft-method to obtain the
            non-wetting fluid configuration.
        'dt'
            Same as 'hybrid', except uses a second distance transform,
            relative to the thresholded mask, to find the invading fluid
            configuration. The choice of 'dt' or 'hybrid' depends on speed,
            which is system and installation specific.
        'mio'
            Using a single morphological image opening step to obtain
            the invading fluid confirguration directly, *then* trims if
            ``access_limitations`` is ``True``. This method is not ideal and
            is included for comparison purposes.

    divs : int or array_like
        The number of times to divide the image for parallel processing.  If ``1``
        then parallel processing does not occur.  ``2`` is equivalent to
        ``[2, 2, 2]`` for a 3D image.  The number of cores used is specified in
        ``porespy.settings.ncores`` and defaults to all cores.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with the pore size values in each voxel.

    See Also
    --------
    porosimetry

    Notes
    -----
    The term *foreground* is used since this function can be applied to
    both pore space or the solid, whichever is set to ``True``.

    This function is identical to ``porosimetry`` with ``access_limited``
    set to ``False``.

    The way local thickness is found in PoreSpy differs from the
    traditional method (i.e. used in ImageJ
    `<https://imagej.net/Local_Thickness>`_). Our approach is probably
    slower, but it allows for the same code to be used for
    ``local_thickness`` and ``porosimetry``, since we can 'trim' invaded
    regions that are not connected to the inlets in the ``porosimetry``
    function. This is not needed in ``local_thickness`` however.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/local_thickness.html>`_
    to view online example.

    """
    im_new = porosimetry(im=im, sizes=sizes, access_limited=False, mode=mode,
                         divs=divs)
    return im_new


def porosimetry(im, sizes=25, inlets=None, access_limited=True, mode='hybrid',
                divs=1):
    r"""
    Performs a porosimetry simulution on an image.

    Parameters
    ----------
    im : ndarray
        An ND image of the porous material containing ``True`` values in the
        pore space.
    sizes : array_like or scalar
        The sizes to invade.  If a list of values of provided they are
        used directly.  If a scalar is provided then that number of points
        spanning the min and max of the distance transform are used.
    inlets : ndarray, boolean
        A boolean mask with ``True`` values indicating where the invasion
        enters the image.  By default all faces are considered inlets,
        akin to a mercury porosimetry experiment.  Users can also apply
        solid boundaries to their image externally before passing it in,
        allowing for complex inlets like circular openings, etc.
        This argument is only used if ``access_limited`` is ``True``.
    access_limited : bool
        This flag indicates if the intrusion should only occur from the
        surfaces (``access_limited`` is ``True``, which is the default),
        or if the invading phase should be allowed to appear in the core
        of the image.  The former simulates experimental tools like
        mercury intrusion porosimetry, while the latter is useful for
        comparison to gauge the extent of shielding effects in the sample.
    mode : str
        Controls with method is used to compute the result. Options are:

        'hybrid'
            (default) Performs a distance tranform of the void
            space, thresholds to find voxels larger than ``sizes[i]``,
            trims the resulting mask if ``access_limitations`` is ``True``,
            then dilates it using the efficient fft-method to obtain the
            non-wetting fluid configuration.
        'dt'
            Same as 'hybrid', except uses a second distance
            transform, relative to the thresholded mask, to find the
            invading fluid configuration. The choice of 'dt' or 'hybrid'
            depends on speed, which is system and installation specific.
        'mio'
            Uses bindary erosion followed by dilation to obtain the invading
            fluid confirguration directly. If ``access_limitations`` is
            ``True`` then disconnected blobs are trimmmed before the dilation.
            This is the only method that can be parallelized by chunking (see
            ``divs`` and ``cores``).

    divs : int or array_like
        The number of times to divide the image for parallel processing.  If ``1``
        then parallel processing does not occur.  ``2`` is equivalent to
        ``[2, 2, 2]`` for a 3D image.  The number of cores used is specified in
        ``porespy.settings.ncores`` and defaults to all cores.

    Returns
    -------
    image : ndarray
        A copy of ``im`` with voxel values indicating the sphere radius at
        which it becomes accessible from the inlets.  This image can be
        used to find invading fluid configurations as a function of
        applied capillary pressure by applying a boolean comparison:
        ``inv_phase = im > r`` where ``r`` is the radius (in voxels) of
        the invading sphere.  Of course, ``r`` can be converted to
        capillary pressure using a preferred model.

    Notes
    -----
    There are many ways to perform this filter, and PoreSpy offers 3,
    which users can choose between via the ``mode`` argument. These
    methods all work in a similar way by finding which foreground voxels
    can accomodate a sphere of a given radius, then repeating for smaller
    radii.

    See Also
    --------
    local_thickness

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/porosimetry.html>`_
    to view online example.

    """
    from porespy.filters import fftmorphology
    im = np.squeeze(im)
    dt = edt(im > 0)

    if inlets is None:
        inlets = get_border(im.shape, mode="faces")

    if isinstance(sizes, int):
        sizes = np.logspace(start=np.log10(np.amax(dt)), stop=0, num=sizes)
    else:
        sizes = np.unique(sizes)[-1::-1]

    if im.ndim == 2:
        strel = ps_disk
        strel_2 = disk
    else:
        strel = ps_ball
        strel_2 = ball

    parallel = False
    if isinstance(divs, int):
        divs = [divs]*im.ndim
    if max(divs) > 1:
        logger.info(f'Performing {insp.currentframe().f_code.co_name} in parallel')
        parallel = True

    if mode == "mio":
        pw = int(np.floor(dt.max()))
        impad = np.pad(im, mode="symmetric", pad_width=pw)
        inlets = np.pad(inlets, mode="symmetric", pad_width=pw)
        # sizes = np.unique(np.around(sizes, decimals=0).astype(int))[-1::-1]
        imresults = np.zeros(np.shape(impad))
        for r in tqdm(sizes, **settings.tqdm):
            if parallel:
                imtemp = chunked_func(func=fftmorphology,
                                      im=impad, strel=strel(r),
                                      overlap=int(r) + 1, mode='erosion',
                                      cores=settings.ncores, divs=divs)
            else:
                imtemp = fftmorphology(im=impad, strel=strel(r), mode='erosion')
            if access_limited:
                imtemp = trim_disconnected_blobs(imtemp, inlets,
                                                 strel=strel_2(1))
            if parallel:
                imtemp = chunked_func(func=fftmorphology,
                                      im=imtemp, strel=strel(r),
                                      overlap=int(r) + 1, mode='dilation',
                                      cores=settings.ncores, divs=divs)
            else:
                imtemp = fftmorphology(im=imtemp, strel=strel(r), mode='dilation')
            if np.any(imtemp):
                imresults[(imresults == 0) * imtemp] = r
        imresults = extract_subsection(imresults, shape=im.shape)
    elif mode == "dt":
        imresults = np.zeros(np.shape(im))
        for r in tqdm(sizes, **settings.tqdm):
            imtemp = dt >= r
            if access_limited:
                imtemp = trim_disconnected_blobs(imtemp, inlets,
                                                 strel=strel_2(1))
            if np.any(imtemp):
                if parallel:
                    imtemp = chunked_func(func=edt,
                                          data=~imtemp, im_arg='data',
                                          overlap=int(r) + 1, parallel=0,
                                          cores=settings.ncores, divs=divs) < r
                else:
                    imtemp = edt(~imtemp) < r
                imresults[(imresults == 0) * imtemp] = r
    elif mode == "hybrid":
        imresults = np.zeros(np.shape(im))
        for r in tqdm(sizes, **settings.tqdm):
            imtemp = dt >= r
            if access_limited:
                imtemp = trim_disconnected_blobs(imtemp, inlets,
                                                 strel=strel_2(1))
            if np.any(imtemp):
                if parallel:
                    imtemp = chunked_func(func=fftmorphology, mode='dilation',
                                          im=imtemp, strel=strel(r),
                                          overlap=int(r) + 1,
                                          cores=settings.ncores, divs=divs)
                else:
                    imtemp = fftmorphology(imtemp, strel(r),
                                           mode="dilation")
                imresults[(imresults == 0) * imtemp] = r
    else:
        raise Exception("Unrecognized mode " + mode)
    return imresults


def trim_disconnected_blobs(im, inlets, strel=None):
    r"""
    Removes foreground voxels not connected to specified inlets.

    Parameters
    ----------
    im : ndarray
        The image containing the blobs to be trimmed
    inlets : ndarray or tuple of indices
        The locations of the inlets.  Can either be a boolean mask the
        same shape as ``im``, or a tuple of indices such as that returned
        by the ``where`` function.  Any voxels *not* connected directly to
        the inlets will be trimmed.
    strel : array-like
        The neighborhood over which connectivity should be checked. It
        must be symmetric and the same dimensionality as the image. It is
        passed directly to the ``scipy.ndimage.label`` function as the
        ``structure`` argument so refer to that docstring for additional
        info.

    Returns
    -------
    image : ndarray
        An array of the same shape as ``im``, but with all foreground
        voxels not connected to the ``inlets`` removed.

    See Also
    --------
    find_disconnected_voxels, find_nonpercolating_paths

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/trim_disconnected_blobs.html>`_
    to view online example.

    """
    if type(inlets) == tuple:
        temp = np.copy(inlets)
        inlets = np.zeros_like(im, dtype=bool)
        inlets[temp] = True
    elif (inlets.shape == im.shape) and (inlets.max() == 1):
        inlets = inlets.astype(bool)
    else:
        raise Exception("inlets not valid, refer to docstring for info")
    if strel is None:
        if im.ndim == 3:
            strel = cube(3)
        else:
            strel = square(3)
    labels = spim.label(inlets + (im > 0), structure=strel)[0]
    keep = np.unique(labels[inlets])
    keep = keep[keep > 0]
    im2 = np.isin(labels, keep)
    im2 = im2 * im
    return im2


def _get_axial_shifts(ndim=2, include_diagonals=False):
    r"""
    Helper function to generate the axial shifts that will be performed on
    the image to identify bordering pixels/voxels
    """
    if ndim == 2:
        if include_diagonals:
            neighbors = square(3)
        else:
            neighbors = diamond(1)
        neighbors[1, 1] = 0
        x, y = np.where(neighbors)
        x -= 1
        y -= 1
        return np.vstack((x, y)).T
    else:
        if include_diagonals:
            neighbors = cube(3)
        else:
            neighbors = octahedron(1)
        neighbors[1, 1, 1] = 0
        x, y, z = np.where(neighbors)
        x -= 1
        y -= 1
        z -= 1
        return np.vstack((x, y, z)).T


def _make_stack(im, include_diagonals=False):
    r"""
    Creates a stack of images with one extra dimension to the input image
    with length equal to the number of borders to search + 1.

    Image is rolled along the axial shifts so that the border pixel is
    overlapping the original pixel. First image in stack is the original.
    Stacking makes direct vectorized array comparisons possible.

    """
    ndim = len(np.shape(im))
    axial_shift = _get_axial_shifts(ndim, include_diagonals)
    if ndim == 2:
        stack = np.zeros([np.shape(im)[0], np.shape(im)[1], len(axial_shift) + 1])
        stack[:, :, 0] = im
        for i in range(len(axial_shift)):
            ax0, ax1 = axial_shift[i]
            temp = np.roll(np.roll(im, ax0, 0), ax1, 1)
            stack[:, :, i + 1] = temp
        return stack
    elif ndim == 3:
        stack = np.zeros(
            [np.shape(im)[0], np.shape(im)[1], np.shape(im)[2], len(axial_shift) + 1]
        )
        stack[:, :, :, 0] = im
        for i in range(len(axial_shift)):
            ax0, ax1, ax2 = axial_shift[i]
            temp = np.roll(np.roll(np.roll(im, ax0, 0), ax1, 1), ax2, 2)
            stack[:, :, :, i + 1] = temp
        return stack


def nphase_border(im, include_diagonals=False):
    r"""
    Identifies the voxels in regions that border *N* other regions.

    Useful for finding triple-phase boundaries.

    Parameters
    ----------
    im : ndarray
        An ND image of the porous material containing discrete values in
        the pore space identifying different regions. e.g. the result of a
        snow-partition
    include_diagonals : bool
        When identifying bordering pixels (2D) and voxels (3D) include
        those shifted along more than one axis

    Returns
    -------
    image : ndarray
        A copy of ``im`` with voxel values equal to the number of uniquely
        different bordering values

    """
    _check_for_singleton_axes(im)
    # Get dimension of image
    ndim = len(np.shape(im))
    if ndim not in [2, 3]:
        raise NotImplementedError("Function only works for 2d and 3d images")
    # Pad image to handle edges
    im = np.pad(im, pad_width=1, mode="edge")
    # Stack rolled images for each neighbor to be inspected
    stack = _make_stack(im, include_diagonals)
    # Sort the stack along the last axis
    stack.sort()
    out = np.ones_like(im)
    # Run through stack recording when neighbor id changes
    # Number of changes is number of unique bordering regions
    for k in range(np.shape(stack)[ndim])[1:]:
        if ndim == 2:
            mask = stack[:, :, k] != stack[:, :, k - 1]
        elif ndim == 3:
            mask = stack[:, :, :, k] != stack[:, :, :, k - 1]
        out += mask
    # Un-pad
    if ndim == 2:
        return out[1:-1, 1:-1].copy()
    else:
        return out[1:-1, 1:-1, 1:-1].copy()


def prune_branches(skel, branch_points=None, iterations=1):
    r"""
    Remove all dangling ends or tails of a skeleton

    Parameters
    ----------
    skel : ndarray
        A image of a full or partial skeleton from which the tails should
        be trimmed.
    branch_points : ndarray, optional
        An image the same size ``skel`` with ``True`` values indicating the
        branch points of the skeleton.  If this is not provided it is
        calculated automatically.
    iterations : int
        The number of times to recursively repeat the process.  The default is
        1.

    Returns
    -------
    array
        An ndarray containing the skeleton with tails removed.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/prune_branches.html>`_
    to view online example.

    """
    skel = skel > 0
    if skel.ndim == 2:
        from skimage.morphology import square as cube
    else:
        from skimage.morphology import cube
    # Create empty image to house results
    im_result = np.zeros_like(skel)
    # If branch points are not supplied, attempt to find them
    if branch_points is None:
        branch_points = spim.convolve(skel * 1.0, weights=cube(3)) > 3
        branch_points = branch_points * skel
    # Store original branch points before dilating
    pts_orig = branch_points
    # Find arcs of skeleton by deleting branch points
    arcs = skel * (~branch_points)
    # Label arcs
    arc_labels = spim.label(arcs, structure=cube(3))[0]
    # Dilate branch points so they overlap with the arcs
    branch_points = spim.binary_dilation(branch_points, structure=cube(3))
    pts_labels = spim.label(branch_points, structure=cube(3))[0]
    # Now scan through each arc to see if it's connected to two branch points
    slices = spim.find_objects(arc_labels)
    label_num = 0
    for s in slices:
        label_num += 1
        # Find branch point labels the overlap current arc
        hits = pts_labels[s] * (arc_labels[s] == label_num)
        # If image contains 2 branch points, then it's not a tail.
        if len(np.unique(hits)) == 3:
            im_result[s] += arc_labels[s] == label_num
    # Add missing branch points back to arc image to make complete skeleton
    im_result += skel * pts_orig
    if iterations > 1:
        iterations -= 1
        im_temp = np.copy(im_result)
        im_result = prune_branches(skel=im_result,
                                   branch_points=None,
                                   iterations=iterations)
        if np.all(im_temp == im_result):
            iterations = 0
    return im_result


def chunked_func(func,
                 overlap=None,
                 divs=2,
                 cores=None,
                 im_arg=["input", "image", "im"],
                 strel_arg=["strel", "structure", "footprint"],
                 **kwargs):
    r"""
    Performs the specfied operation "chunk-wise" in parallel using ``dask``.

    This can be used to save memory by doing one chunk at a time
    (``cores=1``) or to increase computation speed by spreading the work
    across multiple cores (e.g. ``cores=8``)

    This function can be used with any operation that applies a
    structuring element of some sort, since this implies that the
    operation is local and can be chunked.

    Parameters
    ----------
    func : function handle
        The function which should be applied to each chunk, such as
        ``spipy.ndimage.binary_dilation``.
    overlap : scalar or list of scalars, optional
        The amount of overlap to include when dividing up the image. This
        value will almost always be the size (i.e. raduis) of the
        structuring element. If not specified then the amount of overlap
        is inferred from the size of the structuring element, in which
        case the ``strel_arg`` must be specified.
    divs : scalar or list of scalars (default = [2, 2, 2])
        The number of chunks to divide the image into in each direction.
        The default is 2 chunks in each direction, resulting in a
        quartering of the image and 8 total chunks (in 3D).  A scalar is
        interpreted as applying to all directions, while a list of scalars
        is interpreted as applying to each individual direction.
    cores : scalar
        The number of cores which should be used.  By default, all cores
        will be used, or as many are needed for the given number of
        chunks, which ever is smaller.
    im_arg : str
        The keyword used by ``func`` for the image to be operated on. By
        default this function will look for ``image``, ``input``, and
        ``im`` which are commonly used by *scipy.ndimage* and *skimage*.
    strel_arg : str
        The keyword used by ``func`` for the structuring element to apply.
        This is only needed if ``overlap`` is not specified. By default
        this function will look for ``strel``, ``structure``, and
        ``footprint`` which are commonly used by *scipy.ndimage* and
        *skimage*.
    kwargs
        All other arguments are passed to ``func`` as keyword arguments.
        Note that PoreSpy will fetch the image from this list of keywords
        using the value provided to ``im_arg``.

    Returns
    -------
    result : ndarray
        An image the same size as the input image, with the specified
        filter applied as though done on a single large image. There
        should be *no* difference.

    Notes
    -----
    This function divides the image into the specified number of chunks,
    but also applies a padding to each chunk to create an overlap with
    neighboring chunks. This way the operation does not have any edge
    artifacts. The amount of padding is usually equal to the radius of the
    structuring element but some functions do not use one, such as the
    distance transform and Gaussian blur.  In these cases the user can
    specify ``overlap``.

    See Also
    --------
    scikit-image.util.apply_parallel

    Examples
    --------
    `Click here
    <https://porespy.org/examples/filters/howtos/chunked_func.html>`_
    to view online example.

    """

    @dask.delayed
    def apply_func(func, **kwargs):
        # Apply function on sub-slice of overall image
        return func(**kwargs)

    # Determine the value for im_arg
    if type(im_arg) == str:
        im_arg = [im_arg]
    for item in im_arg:
        if item in kwargs.keys():
            im = kwargs[item]
            im_arg = item
            break
    # Fetch image from the kwargs dict
    im = kwargs[im_arg]
    # Determine the number of divisions to create
    divs = np.ones((im.ndim,), dtype=int) * np.array(divs)
    if cores is None:
        cores = settings.ncores
    # If overlap given then use it, otherwise search for strel in kwargs
    if overlap is not None:
        overlap = overlap * (divs > 1)
    else:
        if type(strel_arg) == str:
            strel_arg = [strel_arg]
        for item in strel_arg:
            if item in kwargs.keys():
                strel = kwargs[item]
                break
        overlap = np.array(strel.shape) * (divs > 1)
    slices = subdivide(im=im, divs=divs, overlap=overlap)
    # Apply func to each subsection of the image
    res = []
    for s in slices:
        # Extract subsection from image and input into kwargs
        kwargs[im_arg] = dask.delayed(np.ascontiguousarray(im[tuple(s)]))
        res.append(apply_func(func=func, **kwargs))
    # Have dask actually compute the function on each subsection in parallel
    # with ProgressBar():
        # ims = dask.compute(res, num_workers=cores)[0]
    ims = dask.compute(res, num_workers=cores)[0]
    # Finally, put the pieces back together into a single master image, im2
    im2 = recombine(ims=ims, slices=slices, overlap=overlap)
    return im2



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection


def bar(results, h='pdf', **kwargs):  # pragma: no cover
    r"""
    Convenience wrapper for matplotlib's ``bar``.

    This automatically:

        * fetches the ``bin_centers``
        * fetches the bin heights from the specified ``h``
        * sets the bin widths
        * sets the edges to black

    Parameters
    ----------
    results : object
        The objects returned by various functions in the
        ``porespy.metrics`` submodule, such as ``chord_length_distribution``.
    h : str
        The value to use for bin heights.  The default is ``pdf``, but
        ``cdf`` is another option. Depending on the function the named-tuple
        may have different options.
    kwargs : keyword arguments
        All other keyword arguments are passed to ``bar``, including
        ``edgecolor`` if you wish to overwrite the default black.

    Returns
    -------
    fig: Matplotlib figure handle
    """
    if 'edgecolor' not in kwargs:
        kwargs['edgecolor'] = 'k'
    fig = plt.bar(x=results.bin_centers, height=getattr(results, h),
                  width=results.bin_widths, **kwargs)
    xlab = [attr for attr in results.__dir__() if not attr.startswith('_')][0]
    plt.xlabel(xlab)
    plt.ylabel(h)
    return fig


def imshow(*im, ind=None, axis=None):  # pragma: no cover
    r"""
    Convenience wrapper for matplotlib's ``imshow``.

    This automatically:

        * slices a 3D image in the middle of the last axis
        * uses a masked array to make 0's white
        * sets the origin to 'lower' so bottom-left corner is [0, 0]
        * disables interpolation

    Parameters
    ----------
    im : ndarray
        The 2D or 3D image (or images) to show.  If 2D then all other
        arguments are ignored.
    ind : int
        The slice to show if ``im`` is 3D.  If not given then the middle of
        the image is used.
    axis : int
        The axis to show if ``im`` is 3D.  If not given, then the last
        axis of the image is used, so an 'lower' slice is shown.

    Note
    ----
    ``im`` can also be a series of unnamed arguments, in which case all
    received images will be shown using ``subplot``.
    """
    if not isinstance(im, tuple):
        im = tuple([im])
    for i, image in enumerate(im):
        if image.ndim == 3:
            if axis is None:
                axis = 2
            if ind is None:
                ind = int(image.shape[axis]/2)
            image = image.take(indices=ind, axis=axis)
        image = np.ma.array(image, mask=image == 0)
        fig = plt.subplot(1, len(im), i+1)
        plt.imshow(image, origin='lower', interpolation='none')
    return fig


def show_mesh(mesh):  # pragma: no cover
    r"""
    Visualizes the mesh of a region as obtained by ``get_mesh`` function in
    the ``metrics`` submodule.

    Parameters
    ----------
    mesh : tuple
        A mesh returned by ``skimage.measure.marching_cubes``

    Returns
    -------
    fig : Matplotlib figure
        A handle to a matplotlib 3D axis
    """
    try:
        verts = mesh.vertices
    except AttributeError:
        verts = mesh.verts

    lim_max = np.amax(verts, axis=0)
    lim_min = np.amin(verts, axis=0)

    # Display resulting triangular mesh using Matplotlib.
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # Fancy indexing: `verts[faces]` to generate a collection of triangles
    mesh = Poly3DCollection(verts[mesh.faces])
    mesh.set_edgecolor('k')

    ax.add_collection3d(mesh)
    ax.set_xlabel("x-axis")
    ax.set_ylabel("y-axis")
    ax.set_zlabel("z-axis")
    ax.set_xlim(lim_min[0], lim_max[0])
    ax.set_ylim(lim_min[1], lim_max[1])
    ax.set_zlim(lim_min[2], lim_max[2])

    return fig



import numpy as np


def unpad(im, pad_width):
    r"""
    Remove padding from a previously padded image given original pad widths

    Parameters
    ----------
    im : ndarray
        The padded image from which padding should be removed
    pad_width : int or list of ints
        The amount of padding previously added to each axis. This should be
        the same as the values used to add original padding. Refer to
        ``numpy.pad`` documentation for more details.

    Notes
    -----
    A use case for this is when using ``skimage.morphology.skeletonize_3d``
    to ensure that the skeleton extends beyond the edges of the image, but the
    padding should be subsequently removed.

    Examples
    --------
    `Click here
    <https://porespy.org/examples/tools/howtos/unpad.html>`_
    to view online example.

    """
    pad_width = np.asarray(pad_width).squeeze()

    if pad_width.ndim == 0:
        new_pad_width = []
        for r in range(0, len(im.shape)):
            new_pad_width.append(pad_width)
        pad_width = np.array(new_pad_width)

    if pad_width.ndim == 1:
        shape = im.shape - pad_width[0] - pad_width[1]
        if shape[0] < 1:
            shape = np.array(im.shape) * shape
        s_im = []
        for dim in range(im.ndim):
            lower_im = pad_width[0]
            upper_im = shape[dim] + pad_width[0]
            s_im.append(slice(int(lower_im), int(upper_im)))

    if pad_width.ndim == 2:
        shape = np.asarray(im.shape)
        s_im = []
        for dim in range(im.ndim):
            shape[dim] = im.shape[dim] - pad_width[dim][0] - pad_width[dim][1]
            lower_im = pad_width[dim][0]
            upper_im = shape[dim] + pad_width[dim][0]
            s_im.append(slice(int(lower_im), int(upper_im)))

    return im[tuple(s_im)]



import numpy as np
from edt import edt
from porespy.tools import get_tqdm
import scipy.ndimage as spim
from porespy.tools import get_border, make_contiguous
from porespy.tools import Results
import numba
from porespy import settings
tqdm = get_tqdm()


def ibip(im, inlets=None, dt=None, maxiter=10000):
    r"""
    Performs invasion percolation on given image using iterative image dilation

    Parameters
    ----------
    im : ND-array
        Boolean array with ``True`` values indicating void voxels
    inlets : ND-array
        Boolean array with ``True`` values indicating where the invading fluid
        is injected from.  If ``None``, all faces will be used.
    dt : ND-array (optional)
        The distance transform of ``im``.  If not provided it will be
        calculated, so supplying it saves time.
    maxiter : scalar
        The number of steps to apply before stopping.  The default is to run
        for 10,000 steps which is almost certain to reach completion if the
        image is smaller than about 250-cubed.

    Returns
    -------
    results : Results object
        A custom object with the following two arrays as attributes:

        'inv_sequence'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the sequence at which it was invaded.

        'inv_size'
            An ndarray the same shape as ``im`` with each voxel labelled by
            the ``inv_size`` at which was filled.

    See Also
    --------
    porosimetry

    """
    # Process the boundary image
    if inlets is None:
        inlets = get_border(shape=im.shape, mode='faces')
    bd = np.copy(inlets > 0)
    if dt is None:  # Find dt if not given
        dt = edt(im)
    dt = dt.astype(int)  # Conert the dt to nearest integer
    # Initialize inv image with -1 in the solid, and 0's in the void
    inv = -1*(~im)
    sizes = -1*(~im)
    scratch = np.copy(bd)
    for step in tqdm(range(1, maxiter), **settings.tqdm):
        pt = _where(bd)
        scratch = np.copy(bd)
        temp = _insert_disks_at_points(im=scratch, coords=pt,
                                       r=1, v=1, smooth=False)
        # Reduce to only the 'new' boundary
        edge = temp*(dt > 0)
        if ~edge.any():
            break
        # Find the maximum value of the dt underlaying the new edge
        r_max = (dt*edge).max()
        # Find all values of the dt with that size
        dt_thresh = dt >= r_max
        # Extract the actual coordinates of the insertion sites
        pt = _where(edge*dt_thresh)
        inv = _insert_disks_at_points(im=inv, coords=pt,
                                      r=r_max, v=step, smooth=True)
        sizes = _insert_disks_at_points(im=sizes, coords=pt,
                                        r=r_max, v=r_max, smooth=True)
        dt, bd = _update_dt_and_bd(dt, bd, pt)
    # Convert inv image so that uninvaded voxels are set to -1 and solid to 0
    temp = inv == 0  # Uninvaded voxels are set to -1 after _ibip
    inv[~im] = 0
    inv[temp] = -1
    inv = make_contiguous(im=inv, mode='symmetric')
    # Deal with invasion sizes similarly
    temp = sizes == 0
    sizes[~im] = 0
    sizes[temp] = -1
    results = Results()
    results.inv_sequence = inv
    results.inv_sizes = sizes
    return results


@numba.jit(nopython=True, parallel=False)
def _where(arr):
    inds = np.where(arr)
    result = np.vstack(inds)
    return result


@numba.jit(nopython=True)
def _update_dt_and_bd(dt, bd, pt):
    if dt.ndim == 2:
        for i in range(pt.shape[1]):
            bd[pt[0, i], pt[1, i]] = True
            dt[pt[0, i], pt[1, i]] = 0
    else:
        for i in range(pt.shape[1]):
            bd[pt[0, i], pt[1, i], pt[2, i]] = True
            dt[pt[0, i], pt[1, i], pt[2, i]] = 0
    return dt, bd


@numba.jit(nopython=True, parallel=False)
def _make_disks(r, smooth=True):  # pragma: no cover
    r"""
    Returns a list of disks from size 0 to ``r``

    Parameters
    ----------
    r : int
        The size of the largest disk to generate
    smooth : bool
        Indicates whether the disks should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    disks : list of ND-arrays
        A list containing the disk images, with the disk of radius R at index
        R of the list, meaning it can be accessed as ``disks[R]``.

    """
    disks = [np.atleast_2d(np.array([]))]
    for val in range(1, r):
        disk = _make_disk(val, smooth)
        disks.append(disk)
    return disks


@numba.jit(nopython=True, parallel=False)
def _make_balls(r, smooth=True):  # pragma: no cover
    r"""
    Returns a list of balls from size 0 to ``r``

    Parameters
    ----------
    r : int
        The size of the largest ball to generate
    smooth : bool
        Indicates whether the balls should include the nibs (``False``) on
        the surface or not (``True``).  The default is ``True``.

    Returns
    -------
    balls : list of ND-arrays
        A list containing the ball images, with the ball of radius R at index
        R of the list, meaning it can be accessed as ``balls[R]``.

    """
    balls = [np.atleast_3d(np.array([]))]
    for val in range(1, r):
        ball = _make_ball(val, smooth)
        balls.append(ball)
    return balls


@numba.jit(nopython=True, parallel=False)
def _insert_disks_at_points(im, coords, r, v, smooth=True):  # pragma: no cover
    r"""
    Insert spheres (or disks) into the given ND-image at given locations

    This function uses numba to accelerate the process, and does not
    overwrite any existing values (i.e. only writes to locations containing
    zeros).

    Parameters
    ----------
    im : ND-array
        The image into which the spheres/disks should be inserted. This is an
        'in-place' operation.
    coords : ND-array
        The center point of each sphere/disk in an array of shape
        ``ndim by npts``
    r : int
        The radius of all the spheres/disks to add. It is assumed that they
        are all the same radius.
    v : scalar
        The value to insert
    smooth : boolean
        If ``True`` (default) then the spheres/disks will not have the litte
        nibs on the surfaces.

    """
    npts = len(coords[0])
    if im.ndim == 2:
        xlim, ylim = im.shape
        s = _make_disk(r, smooth)
        for i in range(npts):
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            if (s[a, b] == 1) and (im[x, y] == 0):
                                im[x, y] = v
    elif im.ndim == 3:
        xlim, ylim, zlim = im.shape
        s = _make_ball(r, smooth)
        for i in range(npts):
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            for c, z in enumerate(range(pt[2]-r, pt[2]+r+1)):
                                if (z >= 0) and (z < zlim):
                                    if (s[a, b, c] == 1) and (im[x, y, z] == 0):
                                        im[x, y, z] = v
    return im


@numba.jit(nopython=True, parallel=False)
def _make_disk(r, smooth=True):  # pragma: no cover
    s = np.zeros((2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            if ((i - r)**2 + (j - r)**2)**0.5 <= thresh:
                s[i, j] = 1
    return s


@numba.jit(nopython=True, parallel=False)
def _make_ball(r, smooth=True):  # pragma: no cover
    s = np.zeros((2*r+1, 2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            for k in range(2*r+1):
                if ((i - r)**2 + (j - r)**2 + (k - r)**2)**0.5 <= thresh:
                    s[i, j, k] = 1
    return s


def find_trapped_regions(seq, outlets=None, bins=25, return_mask=True):
    r"""
    Find the trapped regions given an invasion sequence image

    Parameters
    ----------
    seq : ND-array
        An image with invasion sequence values in each voxel.  Regions
        labelled -1 are considered uninvaded, and regions labelled 0 are
        considered solid. Such an image is returned from the
        ``invaded_region`` function.
    outlets : ND-array
        An image the same size as ``seq`` with ``True`` indicating outlets
        and ``False`` elsewhere.  If not given then all image boundaries
        are considered outlets.
    bins : int
        The resolution to use when thresholding the ``seq`` image.  By default
        the invasion sequence will be broken into 25 discrete steps and
        trapping will be identified at each step. A higher value of ``bins``
        will provide a more accurate trapping analysis, but is more time
        consuming. If ``None`` is specified, then *all* the steps will
        analyzed, providing the highest accuracy.
    return_mask : bool
        If ``True`` (default) then the returned image is a boolean mask
        indicating which voxels are trapped.  If ``False``, then a copy of
        ``seq`` is returned with the trapped voxels set to uninvaded and
        the invasion sequence values adjusted accordingly.

    Returns
    -------
    trapped : ND-image
        An image, the same size as ``seq``.  If ``return_mask`` is ``True``,
        then the image has ``True`` values indicating the trapped voxels.  If
        ``return_mask`` is ``False``, then a copy of ``seq`` is returned with
        trapped voxels set to 0.

    """
    if seq.max() <= 1.0:
        raise Exception('seq appears to be saturations, not invasion steps')
    seq = np.copy(seq)
    if outlets is None:
        outlets = get_border(seq.shape, mode='faces')
    trapped = np.zeros_like(outlets)
    if bins is None:
        bins = np.unique(seq)
        bins = bins[bins > 0]
    else:
        bins = np.linspace(seq.max(), 1, bins)
    for i in tqdm(bins, **settings.tqdm):
        temp = seq > i
        labels = spim.label(temp)[0]
        keep = np.unique(labels[outlets])[1:]
        trapped += temp*np.isin(labels, keep, invert=True)
    if return_mask:
        return trapped
    else:
        seq[trapped] = -1
        seq = make_contiguous(seq, mode='symmetric')
        return seq



import numpy as np
from porespy.tools import make_contiguous
from scipy.stats import rankdata


def size_to_seq(size, im=None, bins=None):
    r"""
    Converts an image of invasion size values into sequence values

    This is meant to accept the output of the ``porosimetry`` function.

    Parameters
    ----------
    size : ndarray
        The image containing invasion size values in each voxel.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``size == 0``.
    bins : array_like or int (optional)
        The bins to use when converting sizes to sequence.  The default is
        to create 1 bin for each unique value in ``size``.  If an **int**
        is supplied it is interpreted as the number of bins between 0 and the
        maximum value in ``size``.  If an array is supplied it is used as
        the bins directly.

    Returns
    -------
    seq : ndarray
        An ndarray the same shape as ``size`` with invasion size values
        replaced by the invasion sequence.  This assumes that the invasion
        process occurs via increasing pressure steps, such as produced by
        the ``porosimetry`` function.

    """
    solid = size == 0
    if bins is None:
        bins = np.unique(size)
    elif isinstance(bins, int):
        bins = np.linspace(0, size.max(), bins)
    vals = np.digitize(size, bins=bins, right=True)
    # Invert the vals so smallest size has largest sequence
    vals = -(vals - vals.max() - 1) * ~solid
    # In case too many bins are given, remove empty ones
    vals = make_contiguous(vals, mode='keep_zeros')
    return vals


def size_to_satn(size, im=None, bins=None):
    r"""
    Converts an image of invasion size values into saturations.

    This is meant to accept the output of the ``porosimetry`` function.

    Parameters
    ----------
    size : ndarray
        The image containing invasion size values in each voxel.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``size == 0``.
    bins : array_like or int (optional)
        The bins to use when converting sizes to saturation.  The default is
        to create 1 bin for each unique value in ``size``.  If an **int**
        is supplied it is interpreted as the number of bins between 0 and the
        maximum value in ``size``.  If an array is supplied it is used as
        the bins directly.

    Returns
    -------
    satn : ndarray
        An ndarray the same size as ``seq`` but with sequence values replaced
        by the fraction of void space invaded at or below the sequence number.
        Solid voxels and uninvaded voxels are represented by 0 and -1,
        respectively.
    """
    if bins is None:
        bins = np.unique(size)
    elif isinstance(bins, int):
        bins = np.linspace(0, size.max(), bins)
    if im is None:
        im = (size != 0)
    void_vol = im.sum()
    satn = -np.ones_like(size, dtype=float)
    for r in bins[-1::-1]:
        hits = (size >= r) * (size > 0)
        temp = hits.sum()/void_vol
        satn[hits * (satn == -1)] = temp
    satn *= (im > 0)
    return satn


def seq_to_satn(seq, im=None):
    r"""
    Converts an image of invasion sequence values to saturation values.

    This is meant to accept the output of the ``ibip`` function.

    Parameters
    ----------
    seq : ndarray
        The image containing invasion sequence values in each voxel.  Solid
        should be indicated as 0's and uninvaded voxels as -1.
    im : ndarray, optional
        A binary image of the porous media, with ``True`` indicating the
        void space and ``False`` indicating the solid phase. If not given
        then it is assumed that the solid is identified as ``seq == 0``.

    Returns
    -------
    satn : ndarray
        An ndarray the same size as ``seq`` but with sequence values replaced
        by the fraction of void space invaded at or below the sequence number.
        Solid voxels and uninvaded voxels are represented by 0 and -1,
        respectively.

    """
    seq = np.copy(seq).astype(int)
    if im is None:
        solid_mask = seq == 0
    else:
        solid_mask = im == 0
    uninvaded_mask = seq == -1
    seq[seq <= 0] = 0
    seq = rankdata(seq, method='dense') - 1
    b = np.bincount(seq)
    if (solid_mask.sum() > 0) or (uninvaded_mask.sum() > 0):
        b[0] = 0
    c = np.cumsum(b)
    seq = np.reshape(seq, solid_mask.shape)
    satn = c[seq]/(seq.size - solid_mask.sum())
    satn[solid_mask] = 0
    satn[uninvaded_mask] = -1
    return satn



r"""

Tools
#####

**Helper Functions**

This module contains a variety of functions for manipulating images in
ways do NOT return a modified version of the original image.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

   tools.align_image_with_openpnm
   tools.bbox_to_slices
   tools.extend_slice
   tools.extract_subsection
   tools.extract_regions
   tools.extract_cylinder
   tools.extract_subsection
   tools.find_outer_region
   tools.get_border
   tools.get_planes
   tools.insert_cylinder
   tools.insert_sphere
   tools.in_hull
   tools.isolate_object
   tools.make_contiguous
   tools.marching_map
   tools.mesh_region
   tools.norm_to_uniform
   tools.overlay
   tools.ps_ball
   tools.ps_disk
   tools.ps_rect
   tools.ps_round
   tools.randomize_colors
   tools.recombine
   tools.subdivide
   tools.unpad

"""

__all__ = [
    "align_image_with_openpnm",
    "bbox_to_slices",
    "extend_slice",
    "extract_cylinder",
    "extract_subsection",
    "extract_regions",
    "find_outer_region",
    "get_border",
    "get_planes",
    "insert_cylinder",
    "insert_sphere",
    "in_hull",
    "isolate_object",
    "make_contiguous",
    "marching_map",
    "mesh_region",
    "norm_to_uniform",
    "overlay",
    "ps_ball",
    "ps_disk",
    "ps_rect",
    "ps_round",
    "randomize_colors",
    "recombine",
    "subdivide",
    "unpad",
    "sanitize_filename",
    "get_tqdm",
    "show_docstring"]

from ._funcs import align_image_with_openpnm
from ._funcs import bbox_to_slices
from ._funcs import extend_slice
from ._funcs import extract_cylinder
from ._funcs import extract_subsection
from ._funcs import extract_regions
from ._funcs import find_outer_region
from ._funcs import get_border
from ._funcs import get_planes
from ._funcs import insert_cylinder
from ._funcs import insert_sphere
from ._funcs import in_hull
from ._funcs import isolate_object
from ._funcs import marching_map
from ._funcs import make_contiguous
from ._funcs import mesh_region
from ._funcs import norm_to_uniform
from ._funcs import overlay
from ._funcs import randomize_colors
from ._funcs import recombine
from ._funcs import ps_ball
from ._funcs import ps_disk
from ._funcs import ps_rect
from ._funcs import ps_round
from ._funcs import subdivide
from ._utils import sanitize_filename
from ._utils import get_tqdm
from ._utils import show_docstring
from ._utils import Results
from ._funcs import _check_for_singleton_axes
from ._unpad import unpad



import numba
import numpy as np
import scipy.ndimage as spim
from edt import edt
from skimage.morphology import disk, ball
from porespy import settings
from porespy.tools import get_tqdm, ps_round
from porespy.filters import trim_disconnected_blobs, fftmorphology
from loguru import logger
tqdm = get_tqdm()


@numba.jit(nopython=True, parallel=False)
def insert_disks_at_points(im, coords, radii, v, smooth=True):
    r"""
    Insert spheres of specified radii into an ndarray at given locations.

    Parameters
    ----------
    im : ndarray
        The image into which the spheres/disks should be inserted. This is an
        'in-place' operation.
    coords : ndarray
        The center point of each sphere/disk in an array of shape
        ``ndim by npts``
    radii : array_like
        The radii of the spheres/disks to add.
    v : scalar
        The value to insert
    smooth : boolean
        If ``True`` (default) then the spheres/disks will not have the litte
        nibs on the surfaces.

    Notes
    -----
    This function uses numba to accelerate the process.

    """
    npts = len(coords[0])
    if im.ndim == 2:
        xlim, ylim = im.shape
        for i in range(npts):
            r = radii[i]
            s = _make_disk(r, smooth)
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            if (s[a, b] == 1):  # and (im[x, y] == 0):
                                im[x, y] = v
    elif im.ndim == 3:
        xlim, ylim, zlim = im.shape
        for i in range(npts):
            r = radii[i]
            s = _make_ball(r, smooth)
            pt = coords[:, i]
            for a, x in enumerate(range(pt[0]-r, pt[0]+r+1)):
                if (x >= 0) and (x < xlim):
                    for b, y in enumerate(range(pt[1]-r, pt[1]+r+1)):
                        if (y >= 0) and (y < ylim):
                            for c, z in enumerate(range(pt[2]-r, pt[2]+r+1)):
                                if (z >= 0) and (z < zlim):
                                    if (s[a, b, c] == 1):  # and (im[x, y, z] == 0):
                                        im[x, y, z] = v
    return im


@numba.jit(nopython=True, parallel=False)
def _make_disk(r, smooth=True):
    r"""
    Generate a strel suitable for use in numba nojit function.

    Numba won't allow calls to skimage strel generators so this function
    makes one, also using njit.
    """
    s = np.zeros((2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            if ((i - r)**2 + (j - r)**2)**0.5 <= thresh:
                s[i, j] = 1
    return s


@numba.jit(nopython=True, parallel=False)
def _make_ball(r, smooth=True):
    r"""
    Generate a strel suitable for use in numba nojit function.

    Numba won't allow calls to skimage strel generators so this function
    makes one, also using njit.
    """
    s = np.zeros((2*r+1, 2*r+1, 2*r+1), dtype=type(r))
    if smooth:
        thresh = r - 0.001
    else:
        thresh = r
    for i in range(2*r+1):
        for j in range(2*r+1):
            for k in range(2*r+1):
                if ((i - r)**2 + (j - r)**2 + (k - r)**2)**0.5 <= thresh:
                    s[i, j, k] = 1
    return s


def pseudo_gravity_packing(im, r, clearance=0, axis=0, maxiter=1000):
    r"""
    Iteratively inserts spheres at the lowest accessible point in an image,
    mimicking a gravity packing.

    Parameters
    ----------
    im : ndarray
        Image with ``True`` values indicating the phase where spheres should be
        inserted. A common option would be a cylindrical plug which would
        result in a tube filled with beads.
    r : int
        The radius of the spheres to be added
    clearance : int (default is 0)
        The amount space to add between neighboring spheres. The value can be
        negative for overlapping spheres, but ``abs(clearance) > r``.
    axis : int (default is 0)
        The axis along which gravity acts.
    maxiter : int (default is 1000)
        The maximum number of spheres to add

    Returns
    -------
    spheres : ndarray
        An image the same size as ``im`` with spheres indicated by ``True``.
        The spheres are only inserted at locations that are accessible
        from the top of the image.

    """
    logger.debug(f'Adding spheres of radius {r}')
    im = np.swapaxes(im, 0, axis)
    im_temp = np.zeros_like(im, dtype=bool)
    r = r - 1
    strel = disk if im.ndim == 2 else ball
    sites = fftmorphology(im == 1, strel=strel(r), mode='erosion')
    inlets = np.zeros_like(im)
    inlets[-(r+1), ...] = True
    sites = trim_disconnected_blobs(im=sites, inlets=inlets)
    x_min = np.where(sites)[0].min()
    n = None
    for n in tqdm(range(maxiter), **settings.tqdm):
        if im.ndim == 2:
            x, y = np.where(sites[x_min:x_min+2*r, ...])
        else:
            x, y, z = np.where(sites[x_min:x_min+2*r, ...])
        if len(x) == 0:
            break
        options = np.where(x == x.min())[0]
        choice = np.random.randint(len(options))
        if im.ndim == 2:
            cen = np.vstack([x[options[choice]] + x_min,
                             y[options[choice]]])
        else:
            cen = np.vstack([x[options[choice]] + x_min,
                             y[options[choice]],
                             z[options[choice]]])
        im_temp = insert_disks_at_points(im_temp, coords=cen,
                                         radii=np.array([r - clearance]), v=True)
        sites = insert_disks_at_points(sites, coords=cen,
                                       radii=np.array([2*r]), v=0)
        x_min += x.min()
    logger.debug(f'A total of {n} spheres were added')
    im_temp = np.swapaxes(im_temp, 0, axis)
    return im_temp


def pseudo_electrostatic_packing(im, r, sites=None,
                                 clearance=0,
                                 protrusion=0,
                                 maxiter=1000):
    r"""
    Iterativley inserts spheres as close to the given sites as possible.

    Parameters
    ----------
    im : ndarray
        Image with ``True`` values indicating the phase where spheres should be
        inserted.
    r : int
        Radius of spheres to insert.
    sites : ndarray (optional)
        An image with ``True`` values indicating the electrostatic attraction points.
        If this is not given then the peaks in the distance transform are used.
    clearance : int (optional, default=0)
        The amount of space to put between each sphere. Negative values are
        acceptable to create overlaps, but ``abs(clearance) < r``.
    protrusion : int (optional, default=0)
        The amount that spheres are allowed to protrude beyond the active phase.
    maxiter : int (optional, default=1000)
        The maximum number of spheres to insert.

    Returns
    -------
    im : ndarray
        An image with inserted spheres indicated by ``True``

    """
    im_temp = np.zeros_like(im, dtype=bool)
    dt_im = edt(im)
    if sites is None:
        dt2 = spim.gaussian_filter(dt_im, sigma=0.5)
        strel = ps_round(r, ndim=im.ndim, smooth=True)
        sites = (spim.maximum_filter(dt2, footprint=strel) == dt2)*im
    dt = edt(sites == 0).astype(int)
    sites = (sites == 0)*(dt_im >= (r - protrusion))
    dtmax = int(dt_im.max()*2)
    dt[~sites] = dtmax
    r = r + clearance
    # Get initial options
    options = np.where(dt == 1)
    for _ in tqdm(range(maxiter), **settings.tqdm):
        hits = dt[options] < dtmax
        if hits.sum() == 0:
            if dt.min() == dtmax:
                break
            options = np.where(dt == dt.min())
            hits = dt[options] < dtmax
        if hits.size == 0:
            break
        choice = np.where(hits)[0][0]
        cen = np.vstack([options[i][choice] for i in range(im.ndim)])
        im_temp = insert_disks_at_points(im_temp, coords=cen,
                                         radii=np.array([r-clearance]), v=True)
        dt = insert_disks_at_points(dt, coords=cen,
                                    radii=np.array([2*r-clearance]), v=int(dtmax))
    return im_temp



import numpy as np
from porespy import settings
from porespy.tools import get_tqdm, Results
tqdm = get_tqdm()


def boxcount(im, bins=10, d_min=1, d_max=None):
    r"""
    Calculates fractal dimension of an image using the tiled box counting
    method [1]_

    Parameters
    ----------
    im : ndarray
        The image of the porous material.
    bins : int or array_like, optional
        The number of box sizes to use. The default is 10 sizes
        logarithmically spaced between 1 and ``min(im.shape)``.
        If an array is provided, this is used directly.

    Returns
    -------
    results
        An object possessing the following attributes:

        size : ndarray
            The box sizes used
        count : ndarray
            The number of boxes of each size that contain both solid and void
        slope : ndarray
            The gradient of ``count``. This has the same number of elements as
            ``count`` and

    References
    ----------
    [1] See Boxcounting on `Wikipedia <https://en.wikipedia.org/wiki/Box_counting>`_

    Examples
    --------
    >>> import porespy as ps
    >>> import matplotlib.pyplot as plt
    >>> dust = ps.generators.random_cantor_dust([1024, 1024], n=8)
    >>> f = ps.metrics.boxcount(dust)
    >>> fig, ax = plt.subplots(1, 2)
    >>> ax[0].loglog(f.size, f.count, 'bo-')
    >>> ax[1].semilogx(f.size, f.slope, 'ro-')
    >>> ax[1].set_ylim([0, 4])

    """
    im = np.array(im, dtype=bool)

    if (len(im.shape) != 2 and len(im.shape) != 3):
        raise Exception('Image must be 2-dimensional or 3-dimensional')

    if isinstance(bins, int):
        Ds = np.unique(np.logspace(1, np.log10(min(im.shape)), bins).astype(int))
    else:
        Ds = np.array(bins).astype(int)

    N = []
    for d in tqdm(Ds, **settings.tqdm):
        result = 0
        for i in range(0, im.shape[0], d):
            for j in range(0, im.shape[1], d):
                if len(im.shape) == 2:
                    temp = im[i:i+d, j:j+d]
                    result += np.any(temp)
                    result -= np.all(temp)
                else:
                    for k in range(0, im.shape[2], d):
                        temp = im[i:i+d, j:j+d, k:k+d]
                        result += np.any(temp)
                        result -= np.all(temp)
        N.append(result)
    slope = -1*np.gradient(np.log(np.array(N)), np.log(Ds))
    data = Results()
    data.size = Ds
    data.count = N
    data.slope = slope
    return data



import numpy as np
import scipy.ndimage as spim
# import matplotlib.pyplot as plt
# from mpl_toolkits.mplot3d.art3d import Poly3DCollection


def show_3D(im):  # pragma: no cover
    r"""
    Rotates a 3D image and creates an angled view for rough 2D visualization.

    Because it rotates the image it can be slow for large images, so is
    mostly meant for rough checking of small prototype images.

    Parameters
    ----------
    im : ndarray
        The 3D array to be viewed from an angle

    Returns
    -------
    image : ndarray
        A 2D veiw of the given 3D image

    Notes
    -----
    Although this is meant to be *quick* it can still take on the order of
    minutes to render very large images.  It uses ``scipy.ndimage.rotate``
    with no interpolation to view the 3D image from an angle, then casts the
    result into a 2D projection.

    """
    im = ~np.copy(im)
    if im.ndim < 3:
        raise Exception('show_3D only applies to 3D images')
    im = spim.rotate(input=im, angle=22.5, axes=[0, 1], order=0)
    im = spim.rotate(input=im, angle=45, axes=[2, 1], order=0)
    im = spim.rotate(input=im, angle=-17, axes=[0, 1], order=0, reshape=False)
    mask = im != 0
    view = np.where(mask.any(axis=2), mask.argmax(axis=2), 0)
    view = view.max() - view
    f = view.max()/5
    view[view == view.max()] = -f
    view = (view + f)**2
    return view


def show_planes(im, spacing=10):  # pragma: no cover
    r"""
    Create a quick montage showing a 3D image in all three directions

    Parameters
    ----------
    im : ndarray
        A 3D image of the porous material
    spacing : int (optional, default=10)
        Controls the amount of space to put between each panel

    Returns
    -------
    image : ndarray
        A 2D array containing the views.  This single image can be viewed using
        ``matplotlib.pyplot.imshow``.

    """
    s = spacing
    if np.squeeze(im.ndim) < 3:
        raise Exception('This view is only necessary for 3D images')
    x, y, z = (np.array(im.shape)/2).astype(int)
    im_xy = im[:, :, z]
    im_xz = im[:, y, :]
    im_yz = np.rot90(im[x, :, :])

    new_x = im_xy.shape[0] + im_yz.shape[0] + s

    new_y = im_xy.shape[1] + im_xz.shape[1] + s

    new_im = np.zeros([new_x + 2*s, new_y + 2*s], dtype=im.dtype)

    # Add xy image to upper left corner
    new_im[s:im_xy.shape[0] + s,
           s:im_xy.shape[1] + s] = im_xy
    # Add xz image to lower left coner
    x_off = im_xy.shape[0] + 2*s
    y_off = im_xy.shape[1] + 2*s
    new_im[s:s + im_xz.shape[0],
           y_off:y_off + im_xz.shape[1]] = im_xz
    new_im[x_off:x_off + im_yz.shape[0],
           s:s + im_yz.shape[1]] = im_yz

    return new_im


def sem(im, axis=0):  # pragma: no cover
    r"""
    Simulates an SEM photograph looking into the porous material.

    Features are colored according to their depth into the image, so
    darker features are further away.

    Parameters
    ----------
    im : array_like
        ndarray of the porous material with the solid phase marked as 1 or
        True
    axis : int
        Specifes the axis along which the camera will point.

    Returns
    -------
    image : ndarray
        A 2D greyscale image suitable for use in matplotlib's ``imshow``
        function.

    """
    im = np.array(~im, dtype=int)
    if axis == 1:
        im = np.transpose(im, axes=[1, 0, 2])
    if axis == 2:
        im = np.transpose(im, axes=[2, 1, 0])
    t = im.shape[0]
    depth = np.reshape(np.arange(0, t), [t, 1, 1])
    im = im*depth
    im = np.amax(im, axis=0)
    return im


def xray(im, axis=0):  # pragma: no cover
    r"""
    Simulates an X-ray radiograph looking through the porous material.

    The resulting image is colored according to the amount of attenuation an
    X-ray would experience, so regions with more solid will appear darker.

    Parameters
    ----------
    im : array_like
        ndarray of the porous material with the solid phase marked as 1 or
        True

    axis : int
        Specifes the axis along which the camera will point.

    Returns
    -------
    image : ndarray
        A 2D greyscale image suitable for use in matplotlib\'s ```imshow```
        function.
    """
    im = np.array(~im, dtype=int)
    if axis == 1:
        im = np.transpose(im, axes=[1, 0, 2])
    if axis == 2:
        im = np.transpose(im, axes=[2, 1, 0])
    im = np.sum(im, axis=0)
    return im



import porespy as ps
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from matplotlib import animation
from copy import copy
from porespy import settings


def set_mpl_style():  # pragma: no cover
    r"""
    Prettifies matplotlib's output by adjusting fonts, markersize etc.
    """
    sfont = 12
    mfont = 12
    lfont = 12

    image_props = {'interpolation': 'none',
                   'cmap': 'viridis'}
    line_props = {'linewidth': 2,
                  'markersize': 8,
                  'markerfacecolor': 'w'}
    font_props = {'size': sfont}
    axes_props = {'titlesize': lfont,
                  'labelsize': mfont,
                  'linewidth': 2,
                  'labelpad': 8}
    xtick_props = {'labelsize': sfont,
                   'top': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    ytick_props = {'labelsize': sfont,
                   'right': True,
                   'direction': 'in',
                   'major.size': 6,
                   'major.width': 2}
    legend_props = {'fontsize': mfont,
                    'frameon': False}
    figure_props = {'titlesize': sfont,
                    'autolayout': True}

    plt.rc('font', **font_props)
    plt.rc('lines', **line_props)
    plt.rc('axes', **axes_props)
    plt.rc('xtick', **xtick_props)
    plt.rc('ytick', **ytick_props)
    plt.rc('legend', **legend_props)
    plt.rc('figure', **figure_props)
    plt.rc('image', **image_props)

    if ps.settings.notebook:
        import IPython
        IPython.display.set_matplotlib_formats('svg')


def satn_to_movie(im, satn, cmap='viridis',
                  c_under='grey', c_over='white',
                  v_under=1e-3, v_over=1.0, fps=10, repeat=True):
    r"""

    Notes
    -----
    To save animation as a file use:
    ``ani.save('image_based_ip.gif', writer='imagemagick', fps=3)``
    """
    # Define nice color map
    cmap = copy(plt.cm.get_cmap(name=cmap))
    cmap.set_over(color=c_over)
    cmap.set_under(color=c_under)

    # Reduce inv_satn image to limited number of values to speed-up movie
    target = np.around(satn, decimals=3)
    seq = np.zeros_like(target)  # Empty image to place frame
    movie = []  # List to append each frame
    fig, ax = plt.subplots(1, 1)
    steps = np.unique(target)[1:]
    with tqdm(steps, **settings.tqdm) as pbar:
        for v in steps:
            pbar.update()
            seq += v*(target == v)
            seq[~im] = target.max() + 10
            frame1 = ax.imshow(seq, vmin=v_under, vmax=v_over,
                               animated=True, cmap=cmap, origin='lower',
                               interpolation='none')
            movie.append([frame1])
    ani = animation.ArtistAnimation(fig, movie, interval=int(1000/fps),
                                    blit=True, repeat=repeat,
                                    repeat_delay=1.0)
    return ani



r"""

DNS
###

**Direct Numerical Simulation**

This module contains routines for performing direct numerical
simulations.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

   dns.tortuosity

"""

__all__ = ["tortuosity"]

from ._funcs import tortuosity



import numpy as np
from scipy.signal import fftconvolve


def fftmorphology(im, strel, mode='opening'):
    r"""
    Perform morphological operations on binary images using fft approach for
    improved performance

    Parameters
    ----------
    im : ndarray
        The binary image on which to perform the morphological operation
    strel : ndarray
        The structuring element to use.  Must have the same dims as ``im``.
    mode : string
        The type of operation to perform.  Options are 'dilation', 'erosion',
        'opening' and 'closing'.

    Returns
    -------
    image : ndarray
        A copy of the image with the specified moropholgical operation applied
        using the fft-based methods available in ``scipy.signal.fftconvolve``.

    Notes
    -----
    This function uses ``scipy.signal.fftconvolve`` which *can* be more than
    10x faster than the standard binary morphology operation in
    ``scipy.ndimage``.  This speed up may not always be realized, depending
    on the scipy distribution used.

    Examples
    --------
    >>> import porespy as ps
    >>> from numpy import array_equal
    >>> import scipy.ndimage as spim
    >>> from skimage.morphology import disk
    >>> im = ps.generators.blobs(shape=[100, 100], porosity=0.8)

    Check that erosion, dilation, opening, and closing are all the same as
    the ``scipy.ndimage`` functions:

    >>> result = ps.filters.fftmorphology(im, strel=disk(5), mode='erosion')
    >>> temp = spim.binary_erosion(im, structure=disk(5))
    >>> array_equal(result, temp)
    True

    >>> result = ps.filters.fftmorphology(im, strel=disk(5), mode='dilation')
    >>> temp = spim.binary_dilation(im, structure=disk(5))
    >>> array_equal(result, temp)
    True

    >>> result = ps.filters.fftmorphology(im, strel=disk(5), mode='opening')
    >>> temp = spim.binary_opening(im, structure=disk(5))
    >>> array_equal(result, temp)
    True

    >>> result = ps.filters.fftmorphology(im, strel=disk(5), mode='closing')
    >>> temp = spim.binary_closing(im, structure=disk(5))
    >>> array_equal(result, temp)
    True

    """

    def erode(im, strel):
        t = fftconvolve(im, strel, mode='same') > (strel.sum() - 0.1)
        return t

    def dilate(im, strel):
        t = fftconvolve(im, strel, mode='same') > 0.1
        return t

    im = np.squeeze(im)

    # Perform erosion and dilation
    # The array must be padded with 0's so it works correctly at edges
    temp = np.pad(array=im, pad_width=1, mode='constant', constant_values=0)
    if mode.startswith('ero'):
        temp = erode(temp, strel)
    if mode.startswith('dila'):
        temp = dilate(temp, strel)

    # Remove padding from resulting image
    if im.ndim == 2:
        result = temp[1:-1, 1:-1]
    elif im.ndim == 3:
        result = temp[1:-1, 1:-1, 1:-1]

    # Perform opening and closing
    if mode.startswith('open'):
        temp = fftmorphology(im=im, strel=strel, mode='erosion')
        result = fftmorphology(im=temp, strel=strel, mode='dilation')
    if mode.startswith('clos'):
        temp = fftmorphology(im=im, strel=strel, mode='dilation')
        result = fftmorphology(im=temp, strel=strel, mode='erosion')

    return result



r"""
#######
PoreSpy
#######

**Porous Media Image Analysis in Python**

----

PoreSpy consists of the following modules:

**generators**: Routines for generating artificial images of porous materials
useful for testing and illustration

**filters**: Functions that accept an image and return an altered image

**metrics**: Tools for quantifying properties of images

**networks**: Tools for obtaining pore network representations of images

**visualization**: Helper functions for creating useful views of the image

**io**: Functions for output image data in various formats for use in common
software

**tools**: Various useful tools for working with images

----

Example Usage
#############

Working with PoreSpy was designed to be a series of function calls, similar to
building a macro in ImageJ or using Matlab.  A sample workflow is as follows:

Create a test image (or load one using ``skimage.io.imread``)

.. code-block:: python

    >>> import porespy as ps
    >>> import numpy as np
    >>> import scipy.ndimage as spim
    >>> np.random.seed(0)  # Set number generator for same image each time
    >>> im = ps.generators.blobs(shape=[250, 250])

Apply a filter to the image using tools from ``scipy.ndimage``:

.. code-block:: python

    >>> dt = spim.distance_transform_edt(im)

Use some filters from PoreSpy:

.. code-block:: python

    >>> peaks = ps.filters.snow_partitioning(im=im, dt=dt)

"""

from .__version__ import __version__
from .tools._utils import Settings as _Settings

settings = _Settings()

from . import tools
from . import filters
from . import metrics
from . import networks
from . import generators
from . import dns
from . import visualization
from . import io

from .visualization import imshow

import numpy as _np
_np.seterr(divide='ignore', invalid='ignore')



import numpy as np
from edt import edt


def cylindrical_plug(shape, r=None, axis=2):
    r"""
    Generates a cylindrical plug suitable for use as a mask on a tomogram

    Parameters
    ----------
    shape : array_like
        The shape of the image to create.  Can be 3D, or 2D in which case
        a circle is returned.
    r : int (optional)
        The diameter of the cylinder to create. If not provided then the
        largest possible radius is used to fit within the image.
    axis : int
        The direction along with the cylinder's axis of rotation should be
        oriented.  The default is 2, which is the z-direction.

    Returns
    -------
    cylinder : ndarray
        A boolean image of the size given by ``shape`` with ``True``'s
        inside the cylinder and ``False``'s outside.

    """
    shape = np.array(shape, dtype=int)
    if len(shape) == 3:
        axes = np.array(list(set([0, 1, 2]).difference(set([axis]))), dtype=int)
        im2d = np.ones(shape=shape[axes])
        im2d[int(shape[axes[0]]/2), int(shape[axes[1]]/2)] = 0
        dt = edt(im2d)
        if r is None:
            r = int(min(shape[axes])/2)
        circ = dt < r
        tile_ax = [1, 1, 1]
        tile_ax[axis] = shape[axis]
        circ = np.expand_dims(circ, axis)
        cyl = np.tile(circ, tile_ax)
    if len(shape) == 2:
        im2d = np.ones(shape=shape)
        im2d[int(shape[0]/2), int(shape[1]/2)] = 0
        dt = edt(im2d)
        if r is None:
            r = int(min(shape[axes])/2)
        cyl = dt <= r
    return cyl



r"""

Networks
########

**Obtain Network Representations**

Contains functions for analysing images as pore networks.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

   networks.add_boundary_regions
   networks.generate_voxel_image
   networks.label_phases
   network.label_boundaries
   networks.map_to_regions
   networks.maximal_ball
   networks.regions_to_network
   networks.snow2

"""

from ._funcs import add_boundary_regions
from ._funcs import generate_voxel_image
from ._funcs import label_phases
from ._funcs import label_boundaries
from ._funcs import map_to_regions
from ._maximal_ball import maximal_ball_wrapper
from ._getnet import regions_to_network
from ._snow2 import snow2
from ._utils import _net_dict
from ._snow2 import _parse_pad_width



import numpy as np
import openpnm as op
from porespy.filters import trim_nonpercolating_paths
from porespy.tools import Results
from loguru import logger
from porespy.generators import faces


def tortuosity(im, axis):
    r"""
    Calculate the tortuosity of image in the specified direction

    Parameters
    ----------
    im : ndarray
        The binary image to analyze with ``True`` indicating phase of interest
    axis : int
        The axis along which to apply boundary conditions

    Returns
    -------
    results : Results objects
        The following values are computed and returned as attributes:

        - tortuosity
            Calculated using the ``effective_porosity`` as
            :math:`\tau = \frac{D_{AB}}{D_{eff}} \cdot \varepsilon`.

        - effective_porosity
            Porosity of the image after applying ``trim_nonpercolating_paths``.
            This removes disconnected voxels which cause singular matrices.

        - original_porosity
            Porosity of the as-received the image

        - formation_factor
            Found as :math:`D_{AB}/D_{eff}`.

        - concentration
            An image containing the concentration values from the
            simulation.

    """
    if axis > (im.ndim - 1):
        raise Exception(f"'axis' must be <= {im.ndim}")

    # Obtain original porosity
    eps0 = im.sum() / im.size
    # removing floating pores
    IN = faces(im.shape, inlet=axis)
    OUT = faces(im.shape, outlet=axis)
    im = trim_nonpercolating_paths(im, inlets=IN, outlets=OUT)
    # porosity is changed because of trimmimg floating pores
    eps = im.sum() / im.size
    if eps < eps0:  # pragma: no cover
        logger.warning(f'True porosity is {eps:.2f}, filled {eps0 - eps:.5f}'
                       ' volume fraction of the image for it to percolate.')
    # cubic network generation
    net = op.network.CubicTemplate(template=im, spacing=1)
    # Adding phase
    water = op.phases.Water(network=net)
    water['throat.diffusive_conductance'] = 1  # dummy value
    # Running Fickian Diffusion
    fd = op.algorithms.FickianDiffusion(network=net, phase=water)
    # Choosing axis of concentration gradient
    inlets = net['pore.coords'][:, axis] <= 1
    outlets = net['pore.coords'][:, axis] >= im.shape[axis]-1
    # Boundary conditions on concentration
    C_in = 1.0
    C_out = 0.0
    fd.set_value_BC(pores=inlets, values=C_in)
    fd.set_value_BC(pores=outlets, values=C_out)
    fd.settings['solver_family'] = 'scipy'
    fd.settings['solver_type'] = 'cg'
    fd.run()
    # Calculating molar flow rate, effective diffusivity and tortuosity
    rate_out = fd.rate(pores=outlets)[0]
    rate_in = fd.rate(pores=inlets)[0]
    if not np.allclose(-rate_out, rate_in):  # pragma: no cover
        raise Exception('Something went wrong, inlet and outlet rate do not match')
    delta_C = C_in - C_out
    L = im.shape[axis]
    A = np.prod(im.shape) / L
    N_A = A / (L-1) * delta_C  # -1 because BCs are put inside domain, see #495
    Deff = rate_in / N_A
    tau = eps / Deff
    result = Results()
    result.tortuosity = tau
    result.formation_factor = 1 / Deff
    result.original_porosity = eps0
    result.effective_porosity = eps
    conc = np.zeros([im.size, ], dtype=float)
    conc[net['pore.template_indices']] = fd['pore.concentration']
    conc = np.reshape(conc, newshape=im.shape)
    result.concentration = conc
    return result



import numpy as np
from skimage.restoration import denoise_nl_means, estimate_sigma
from skimage.exposure import rescale_intensity, match_histograms
import dask
dask.config.set(scheduler="threads")
# from dask.diagnostics import ProgressBar


def nl_means_layered(im, cores=None, axis=0, patch_size=5, patch_distance=15, h=4):
    r"""
    Apply the non-local means filter to each 2D layer of a stack in parallel.

    This applies ``skimage.restoration.denoise_nl_means`` to each slice, so
    refer to the documentation of that function for further information.

    Parameters
    ----------
    im : ndarray
        The greyscale image with noise to be removed
    cores : int (optional)
        The number of cores to use for the processing. By default all
        available cores are used.
    axis : int
        The axis along which slices should be taken. This should
        correspond to the axis of rotation of the tomography stage, so if
        the sample was rotated about the z-axis, then use ``axis=2``.
    patch_size : int
        Size of patches used for denoising
    patch_distance : int
        Maximal distance in pixels where to search patches used for
        denoising.
    h : float
        Cut-off distance (in gray levels). The higher ``h``, the more
        permissive one is in accepting patches. A higher h results in a
        smoother image, at the expense of blurring features. For a
        Gaussian noise of standard deviation sigma, a rule of thumb is to
        choose the value of ``h`` to be sigma of slightly less.

    Notes
    -----
    The quality of the result depends on ``patch_size``,
    ``patch_distance``, ``h``, and ``sigma``.  It is recommended to
    experiment with a single slice first until a suitable set of
    parameters is found.

    Each slice in the stack is adjusted to have the same histogram and
    intensity.

    """

    @dask.delayed
    def apply_func(func, **kwargs):
        return func(**kwargs)

    temp = np.copy(im)
    for i in range(im.shape[2]):
        temp[:, :, i] = match_histograms(temp[:, :, i], temp[:, :, 0],
                                         multichannel=False)
    p2, p98 = np.percentile(temp, (2, 98))
    temp = rescale_intensity(temp, in_range=(p2, p98))
    temp = temp / temp.max()
    sigma_est = np.mean(estimate_sigma(temp[:, :, 0], multichannel=False))

    kw = {'image': temp,
          'patch_size': patch_size,
          'patch_distance': patch_distance,
          'h': h * sigma_est,
          'multichannel': False,
          'fast_mode': True}

    temp = np.swapaxes(temp, 0, axis)
    results = []
    for i in range(im.shape[2]):
        layer = temp[i, ...]
        kw["image"] = layer
        t = apply_func(func=denoise_nl_means, **kw)
        results.append(t)
    # with ProgressBar():
    #     ims = dask.compute(results, num_workers=cores)[0]
    ims = dask.compute(results, num_workers=cores)[0]
    result = np.array(ims)
    result = np.swapaxes(result, 0, axis)
    return result
